"""
$Id: ip.py 2660 2021-06-02 10:00:04Z pe $

EBAS Masterdata Class for instrument/parameter validity
"""

from .offline_masterdata import IPOfflineMasterData
from .base import EbasMasterBase
from .pg_pl import EbasMasterPL
from .co import PTR_MS_COMP

def _exceptions():
    """
    Generate dictionary for exceptions (special datalevels, not imported but
    should be just file format checked.
    Parameters:
        None
    Returns:
        exception dict
    """
    exceptions = {'0': {}, '1': {}}

    # ACSM lev0:
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm1_non_refractory',
        'collection_efficiency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm1_non_refractory',
            'CO_COMP_NAME': 'collection_efficiency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm1_non_refractory',
        'relative_ionization_efficiency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm1_non_refractory',
            'CO_COMP_NAME': 'relative_ionization_efficiency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm1_non_refractory',
        'nitrogen_ion_flow_signal')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm1_non_refractory',
            'CO_COMP_NAME': 'nitrogen_ion_flow_signal'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm1_non_refractory',
        'ionization_efficiency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm1_non_refractory',
            'CO_COMP_NAME': 'ionization_efficiency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm25_non_refractory',
        'collection_efficiency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm25_non_refractory',
            'CO_COMP_NAME': 'collection_efficiency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm25_non_refractory',
        'relative_ionization_efficiency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm25_non_refractory',
            'CO_COMP_NAME': 'relative_ionization_efficiency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm25_non_refractory',
        'nitrogen_ion_flow_signal')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm25_non_refractory',
            'CO_COMP_NAME': 'nitrogen_ion_flow_signal'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'pm25_non_refractory',
        'ionization_efficiency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'pm25_non_refractory',
            'CO_COMP_NAME': 'ionization_efficiency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'pressure')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'pressure'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'temperature')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'temperature'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'relative_humidity')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'relative_humidity'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'pressure')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'pressure'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'temperature')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'temperature'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'relative_humidity')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'relative_humidity'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'electric_tension')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'electric_tension'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'electric_current')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'electric_current'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'electric_power')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'electric_power'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'airbeam_signal')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'airbeam_signal'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'flow_rate')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'flow_rate'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'frequency')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'frequency'
        }
    exceptions['0'][('aerosol_mass_spectrometer', 'IMG', 'instrument',
        'status')] = {
            'FT_TYPE': 'aerosol_mass_spectrometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'status'
        }
    # NOX lev 0
    basedict = {
        'RE_REGIME_CODE': 'IMG',
        'MA_MATRIX_NAME': 'air',
    }
    for ft_ in ('chemiluminescence_photolytic',
                'chemiluminescence_molybdenum',
                'chemiluminescence_photometer'):
        for co_ in ('NO_#counts', 'NO_converter_#counts',
                    'NO_sensitivity', 'converter_efficiency'):
            dict_ = basedict.copy()
            dict_.update({'FT_TYPE': ft_, 'CO_COMP_NAME': co_})
            exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
                 dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_
        basedict['MA_MATRIX_NAME'] = 'instrument'
        dict_ = basedict.copy()
        dict_.update({'FT_TYPE': ft_, 'CO_COMP_NAME': 'status'})
        exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
             dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_


    # xMPS lev0:
    basedict = {
        'RE_REGIME_CODE': 'IMG',
    }
    for ft_ in ('dmps', 'smps', 'tdmps', 'tsmps', 'v-smps', 'v-tdmps'):
        for comb in (('flow_rate', ('instrument',)),
                     ('status', ('instrument',)),
                     ('particle_number_concentration',
                         ('pm1', 'pm10', 'aerosol',)),
                     ('particle_diameter', ('pm1', 'pm10', 'aerosol'))):
            for ma_ in comb[1]:
                dict_ = basedict.copy()
                dict_.update({'FT_TYPE': ft_, 'CO_COMP_NAME': comb[0],
                              'MA_MATRIX_NAME': ma_})
                exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
                     dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_
    for ft_ in ('v-smps', 'v-tdmps'):
        for comb in (('flow_rate', ('instrument',)),
                     ('status', ('instrument',)),
                     ('particle_number_concentration',
                         ('pm10_non_volatile',)),
                     ('particle_diameter', ('pm10_non_volatile',))):
            for ma_ in comb[1]:
                dict_ = basedict.copy()
                dict_.update({'FT_TYPE': ft_, 'CO_COMP_NAME': comb[0],
                              'MA_MATRIX_NAME': ma_})
                exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
                     dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_

    # filter_absorption_photometer lev 0 (AE33):
    basedict = {
        'RE_REGIME_CODE': 'IMG',
        'FT_TYPE': 'filter_absorption_photometer',
    }
    for comb in (('flow_rate', ('instrument',)),
                 ('status', ('instrument',)),
                 ('filter_number', ('instrument',)),
                 ('sample_length_on_filter', ('instrument',)),
                 ('equivalent_black_carbon_loading', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('biomass_burning_aerosol_fraction', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('filter_loading_compensation_parameter', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('reference_beam_signal', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('sensing_beam_signal', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('transmittance', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('sample_intensity', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ('reference_intensity', ('pm1', 'pm10', 'pm25', 'aerosol')),
                 ):
        for ma_ in comb[1]:
            dict_ = basedict.copy()
            dict_.update({'CO_COMP_NAME': comb[0], 'MA_MATRIX_NAME': ma_})
            exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
                 dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_

    # CCNC lev0:
    basedict = {
        'RE_REGIME_CODE': 'IMG',
        'MA_MATRIX_NAME': 'instrument',
    }
    for comp in ('supersaturation', 'temperature_gradient', 'flow_rate',
                 'electric_current', 'electric_tension', 'status'):
        for inst in ('CCNC', 'DMPS-CCNC'):
            dict_ = basedict.copy()
            dict_.update({'CO_COMP_NAME': comp, 'FT_TYPE': inst})
            exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
                 dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_
    # DMPS-CCNC only:
    basedict = {
        'RE_REGIME_CODE': 'IMG',
        'MA_MATRIX_NAME': 'pm10',
    }
    for comp in ('particle_diameter',
                 'cloud_condensation_nuclei_number_concentration'):
        dict_ = basedict.copy()
        dict_.update({'CO_COMP_NAME': comp, 'FT_TYPE': 'DMPS-CCNC'})
        exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
             dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_

    # CCNC lev1:
    basedict = {
        'RE_REGIME_CODE': 'IMG',
        'MA_MATRIX_NAME': 'instrument',
        'CO_COMP_NAME': 'supersaturation',
    }
    for ft_ in ('CCNC', 'DMPS-CCNC'):
        dict_ = basedict.copy()
        dict_.update({'FT_TYPE': ft_})
        exceptions['1'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
             dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_

    # nephelometer lev0
    basedict = {
        'RE_REGIME_CODE': 'IMG',
        'FT_TYPE': 'nephelometer',
    }
    for comb in (
        ('flow_rate', ('instrument',)),
        ('electric_tension', ('instrument',)),
        ('electric_current', ('instrument',)),
        ('status', ('instrument',)),
        ('aerosol_light_scattering_coefficient_zero_measurement',
         ('pm1', 'pm10', 'aerosol',)),
        ('aerosol_light_backscattering_coefficient_zero_measurement',
         ('pm1', 'pm10', 'aerosol',)),
        ('aerosol_light_rayleighscattering_coefficient_zero_measurement',
         ('pm1', 'pm10', 'aerosol',)),
       ):
        for ma_ in comb[1]:
            dict_ = basedict.copy()
            dict_.update({'CO_COMP_NAME': comb[0], 'MA_MATRIX_NAME': ma_})
            exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
                 dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_

    # cpc lev0:
    basedict = {
        'FT_TYPE': 'cpc',
        'RE_REGIME_CODE': 'IMG',
        'MA_MATRIX_NAME': 'instrument',
    }
    for comp in ('flow_rate', 'pulse_width'):
        dict_ = basedict.copy()
        dict_['CO_COMP_NAME'] = comp
        exceptions['0'][(dict_['FT_TYPE'], dict_['RE_REGIME_CODE'],
             dict_['MA_MATRIX_NAME'], dict_['CO_COMP_NAME'])] = dict_

    # VOC lev0:
    pl_ = EbasMasterPL()
    for comp in [x.CO_COMP_NAME for x in pl_.META['VOC']]:
        for ext in ('_peak_area', '_peak_width', '_retention_time'):
            compmod = comp + ext
            exceptions['0'][('online_gc', 'IMG', 'air', compmod)] = {
                'FT_TYPE': 'online_gc',
                'RE_REGIME_CODE': 'IMG',
                'MA_MATRIX_NAME': 'air',
                'CO_COMP_NAME': compmod,
            }
    exceptions['0'][('online_gc', 'IMG', 'instrument', 'status')] = {
        'FT_TYPE': 'online_gc',
        'RE_REGIME_CODE': 'IMG',
        'MA_MATRIX_NAME': 'air',
        'CO_COMP_NAME': 'status',
    }

    # PTR-MS lev0:
    for comp in PTR_MS_COMP:
        exceptions['0'][('PTR-MS', 'IMG', 'air', comp)] = {
            'FT_TYPE': 'PTR-MS',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'air',
            'CO_COMP_NAME': comp,
        }
    for comp in ('status', 'temperature', 'pressure', 'electric_tension'):
        exceptions['0'][('PTR-MS', 'IMG', 'instrument', comp)] = {
            'FT_TYPE': 'PTR-MS',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': comp,
        }
    # cloud_light_scattering_photometer
    exceptions['0'][('cloud_light_scattering_photometer', 'IMG', 'instrument',
        'electric_tension')] = {
            'FT_TYPE': 'cloud_light_scattering_photometer',
            'RE_REGIME_CODE': 'IMG',
            'MA_MATRIX_NAME': 'instrument',
            'CO_COMP_NAME': 'electric_tension',
        }

    # Finally:
    return exceptions

class EbasMasterIP(EbasMasterBase, IPOfflineMasterData):
    """
    Domain Class for Parameter masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling parameter and checking them against master data.
    Parameter master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for IP (parameters)
    # Those are fallback values, will be read from database as soon as possible.
    IPOfflineMasterData.read_pickle_file()
    # some hardcoded exceptional parameters (see method exceptional):
    IP_EXCEPTIONAL = _exceptions()

    def __init__(self, dbh=None):
        """
        Read ip masterdata if dbh is provided.
        """
        IPOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    def __getitem__(self, key):
        """
        Allows dictionary like access to metadata.
        Exception for IP:
            - exceptional metadata lookup (non DB, data level dependent)
        Parameters:
            key    (instr_type, regime, matrix, comp_name, data_level)
                   Only the first 4 are used for regular lookup!
                   data_level is optional and is used for finding exceptional
                   data level dependent masterdata
        """
        # first try: use regular masterdata, ignore data level
        try:
            return self.__class__.META[key[:4]]
        except KeyError:
            # don't exist: try exceptional masterdata (needs data level)
            # Those masterdata are NOT defined in the database, but might be
            # used e.g. in lev 0 files.
            # Thus they are accepted when reading the file, but the domain
            # layer will issue an error message.
            if len(key) != 5:
                raise
            data_level = key[4]
            if data_level in ('0a', '0b'):
                data_level = '0'
            return self.__class__.IP_EXCEPTIONAL[data_level][key[:4]]



"""
$Id: read_filter.py 2759 2021-11-30 11:41:16Z pe $

File input functionality for EbasEanetFilter
"""

import datetime
from nilutility.datatypes import DataObject, recursive_data_object
from fileformats.EANET import EanetFilter, EanetError
from .eanet_cfg import EANET_SITES, EANET_FILTER_GLOBAL, EANET_FILTER_PARAMS, \
    EANET_FILTER_DETAIL
from .base import EbasEanetFilterBase, EbasEanetReadError

EANET_SITES = recursive_data_object(EANET_SITES)
EANET_FILTER_GLOBAL = recursive_data_object(EANET_FILTER_GLOBAL)
EANET_FILTER_PARAMS = recursive_data_object(EANET_FILTER_PARAMS)
EANET_FILTER_DETAIL = recursive_data_object(EANET_FILTER_DETAIL)


class EanetFilterPartialRead(EbasEanetFilterBase):
    """
    Read fuctionality for EbasEanetFilter I/O object.
    """

    UNIQUE_ID = 0

    def get_unique_id(self):
        """
        Get a unique int id
        """
        self.__class__.UNIQUE_ID += 1
        return self.__class__.UNIQUE_ID

    def _read(self, filename, skip_unitconvert=False, **args):
        """
        Reads a EANET filter file into an EBAS I/O object.
        Parameters:
            filename          path and filename
            other paramters are ignored: ignore_rescode, ignore revdate,
                ignore_parameters, skip_data, skip_unitconvert, skip_variables,
                encoding, ignore_numformat

        Returns:
            None
        Raises:
            IOError (from builtin, file open)
            EbasEanetReadError
        """
        # EANET filter file has no revdata metadata, use now:
        self.metadata.revdate = datetime.datetime.utcnow()
        self.sourcefile = EanetFilter()
        try:
            self.sourcefile.read(filename)
        except EanetError:
            pass
        self.errors += self.sourcefile.errors
        self.warnings += self.sourcefile.warnings
        if self.errors:
            self.logger.info("Exiting because of previous errors")
            raise EbasEanetReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))
        self.parse()
        self.prepare_write(False)

    def parse(self):
        """
        Parse all metadata and data from the source file object to the ebas-io
        object.
        """
        global_meta = EANET_FILTER_GLOBAL.EBAS.copy()
        global_eanet_meta = DataObject()
        global_eanet_meta.update(EANET_FILTER_GLOBAL.EANET)
        try:
            global_eanet_meta.update(
                EANET_SITES[self.sourcefile.site_code]['EANET'])
            global_meta.update(EANET_SITES[self.sourcefile.site_code]['EBAS'])
        except KeyError:
            self.error("Site '{}' not found in configuration".format(
                self.sourcefile.site_code))
            raise EbasEanetReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))

        if self.sourcefile.site_name != global_eanet_meta.site_name:
            self.error(
                "Site name mismatch: file '{}' inconsistent with "
                "configuration".format(self.sourcefile.file_name))
        self.sample_times = self.sourcefile.sample_times
        for var in self.sourcefile.variables:
            metadata = global_meta.copy()
            if not [x for x in var.samples if x is not None]:
                self.warning(
                    "Parameter: '{}': no valid values; skipping".format(
                        var.name))
                continue
            eanet_meta = global_eanet_meta.copy()
            try:
                eanet_meta.update(EANET_FILTER_PARAMS[var.name]['EANET'])
                metadata.update(EANET_FILTER_PARAMS[var.name]['EBAS'])
            except KeyError:
                self.error(
                    "Parameter '{}' not found in configuration".format(
                        var.name))
                continue

            # Update/override metadata from EANET_PRECIP_DETAIL configuration
            try:
                EANET_FILTER_DETAIL[self.sourcefile.site_code]
            except KeyError:
                self.error(
                    "Site '{}' not found in EANET_FILTER_DETAIL "
                    "configuration".format(self.sourcefile.site_code))
                raise EbasEanetReadError(
                    "{} Errors, {} Warnings".format(self.errors,
                                                    self.warnings))
            else:
                try:
                    cfg = EANET_FILTER_DETAIL[
                        self.sourcefile.site_code][var.name]
                except KeyError:
                    try:
                        cfg = EANET_FILTER_DETAIL[
                            self.sourcefile.site_code][None]
                    except KeyError:
                        self.error(
                            "Configuration error {} / {}/None".format(
                                self.sourcefile.site_code, var.name))
                        continue
                eanet_meta.update(cfg.EANET)
                metadata.update(cfg.EBAS)

            self._set_station(metadata)
            self._set_org(metadata)
            self._set_method(metadata)

            if var.unit != eanet_meta.unit:
                self.error(
                    "Parameter: '{}': EANET unit mismatch between file and "
                    "configuration".format(var.name))
            self._add_variable_nondomain(
                self.get_unique_id(), metadata, var.samples, var.flags, 0)



"""
ebas/domain/masterdata/co.py
$Id: co.py 2660 2021-06-02 10:00:04Z pe $

EBAS Masterdata Class for component masterdata

This module implements the class EbasMasterCO.

History:
V.1.0.0  2014-02-25  pe  initial version

"""

from __future__ import absolute_import
import re
from .offline_masterdata import COOfflineMasterData
from .base import EbasMasterBase
from .pg_pl import EbasMasterPL

PTR_MS_COMP = (
    # not only organics
    'mass_21_species_count_rate', 'mass_25_species_count_rate',
    'mass_30_species_count_rate', 'mass_32_species_count_rate',
    'mass_37_species_count_rate', 'mass_55_species_count_rate',
    # organics by m/z (may be different comps)
    'mass_47_organic_compounds_count_rate',
    'mass_57_organic_compounds_count_rate',
    'mass_59_organic_compounds_count_rate',
    'mass_61_organic_compounds_count_rate',
    'mass_69_organic_compounds_count_rate',
    'mass_71_organic_compounds_count_rate',
    'mass_73_organic_compounds_count_rate',
    'mass_93_organic_compounds_count_rate',
    'mass_107_organic_compounds_count_rate',
    'mass_121_organic_compounds_count_rate',
    # specific organic compounds
    'methanal_count_rate', 'methanol_count_rate', 'acetonitrile_count_rate',
    'ethanal_count_rate', 'dimethylsulfide_count_rate',
    'methyl_acetate_count_rate', 'benzene_count_rate', 'styrene_count_rate',
    'chlorobenzene_count_rate', 'monoterpenes_count_rate',
    # NEW styrene, chlorobenzene
    # ToF mass_47_organic_compounds_count_rate:
    'formic_acid_count_rate', 'ethanol_count_rate',
    # ToF mass_61_organic_compounds_count_rate:
    'mass_61.028_organic_compounds_count_rate',
    'mass_61.065_organic_compounds_count_rate',
    # ToF mass_69_organic_compounds_count_rate:
    # 'furan_count_rate',  # TODO: furan not in regular components
    'isoprene_count_rate',
    # ToF mass_73_organic_compounds_count_rate:
    'mass_73.028_organic_compounds_count_rate',
    'mass_73.065_organic_compounds_count_rate',
    # ToF mass_93_organic_compounds_count_rate:
    # 'chloroacetone_count_rate',  # TODO: chloroacetone not in regular components
    # 'epoxybenzene_count_rate',  # TODO: epoxybenzene not in regular components
    'mass_93.06_organic_compounds_count_rate',
    'mass_93.070_organic_compounds_count_rate',
    'mass_93.091_organic_compounds_count_rate',
    # ToF mass_107_organic_compounds_count_rate:
    'benzaldehyde_count_rate',
    'mass_107.086_organic_compounds_count_rate')

class EbasMasterCO(EbasMasterBase, COOfflineMasterData):
    """
    Domain Class for components masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling components and checking them against master data.
    Components master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for CO (components)
    # Those are fallback values, will be read from database as soon as possible.
    COOfflineMasterData.read_pickle_file()
    # some hardcoded exceptional components (see method exceptional):
    CO_EXCEPTIONAL = {
        "0": {
            # ACSM lev0:
            'collection_efficiency': {
                'CO_COMP_NAME': 'collection_efficiency',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            'nitrogen_ion_flow_signal': {
                'CO_COMP_NAME': 'nitrogen_ion_flow_signal',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            'relative_ionization_efficiency': {
                'CO_COMP_NAME': 'relative_ionization_efficiency',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            'airbeam_signal': {
                'CO_COMP_NAME': 'airbeam_signal',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            'frequency': {
                'CO_COMP_NAME': 'frequency',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            'electric_power': {
                'CO_COMP_NAME': 'electric_power',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            'ionization_efficiency': {
                'CO_COMP_NAME': 'ionization_efficiency',
                'CO_CAPTION': None,
                'CO_DESC': None,
            },
            # NOX lev 0:
            'NO_#counts': {
                'CO_COMP_NAME': 'NO_#counts',
                'CO_CAPTION': 'NO_count',
                'CO_DESC': None,
            },
            'NO_converter_#counts':{
                'CO_COMP_NAME': 'NO_converter_#counts',
                'CO_CAPTION': 'NOc_count',
                'CO_DESC': None,
            },
            'NO_sensitivity':{
                'CO_COMP_NAME': 'NO_sensitivity',
                'CO_CAPTION': 'NO_sens',
                'CO_DESC': None,
            },
            'converter_efficiency':{
                'CO_COMP_NAME': 'converter_efficiency',
                'CO_CAPTION': 'converter_eff',
                'CO_DESC': None,
            },

            # dmps lev 0:
            'particle_diameter': {
                'CO_COMP_NAME': 'particle_diameter',
                'CO_CAPTION': 'p_diam',
                'CO_DESC': None,
            },
            # filter_absorption_photometer lev 0:
            'equivalent_black_carbon_loading': {
                'CO_COMP_NAME': 'equivalent_black_carbon_loading',
                'CO_CAPTION': 'BCmass',
                'CO_DESC': None,
            },
            # filter_absorption_photometer lev 0 (clap):
            'sample_length_on_filter': {
                'CO_COMP_NAME': 'sample_length_on_filter',
                'CO_CAPTION': 'smpllen',
                'CO_DESC': None,
            },
            # filter_absorption_photometer lev 0 (AE33):
            'filter_number': {
                'CO_COMP_NAME': 'filter_number',
                'CO_CAPTION': 'tpcnt',
                'CO_DESC': None,
            },
            # filter_absorption_photometer lev 0 (AE33):
            'biomass_burning_aerosol_fraction': {
                'CO_COMP_NAME': 'biomass_burning_aerosol_fraction',
                'CO_CAPTION': 'BB',
                'CO_DESC': None,
            },
            # filter_absorption_photometer lev 0 (AE33):
            'filter_loading_compensation_parameter': {
                'CO_COMP_NAME': 'filter_loading_compensation_parameter',
                'CO_CAPTION': 'k',
                'CO_DESC': None,
            },
            # nephelometer lev0
            'aerosol_light_scattering_coefficient_zero_measurement': {
                'CO_COMP_NAME':
                    'aerosol_light_scattering_coefficient_zero_measurement',
                'CO_CAPTION': 'scat0',
                'CO_DESC': None,
            },
            'aerosol_light_backscattering_coefficient_zero_measurement': {
                'CO_COMP_NAME':
                    'aerosol_light_backscattering_coefficient_zero_measurement',
                'CO_CAPTION': 'bscat0',
                'CO_DESC': None,
            },
            'aerosol_light_rayleighscattering_coefficient_zero_measurement': {
                'CO_COMP_NAME':
                    'aerosol_light_rayleighscattering_coefficient_'
                    'zero_measurement',
                'CO_CAPTION': 'rscat0',
                'CO_DESC': None,
            },
            # CCNC, DMPS-CCNC lev0:
            'supersaturation': {
                'CO_COMP_NAME': 'supersaturation',
                'CO_CAPTION': 'SS',
                'CO_DESC': None,
            },
            'temperature_gradient': {
                'CO_COMP_NAME': 'temperature_gradient',
                'CO_CAPTION': 'dT',
                'CO_DESC': None,
            },
            # cpc lev0:
            'pulse_width': {
                'CO_COMP_NAME': 'pulse_width',
                'CO_CAPTION': 'pulse',
                'CO_DESC': None,
            },
        },
        "1": {
            # CCNC, DMPS-CCNC lev1:
            'supersaturation': {
                'CO_COMP_NAME': 'supersaturation',
                'CO_CAPTION': 'SS',
                'CO_DESC': None,
            },
        },
    }
    # VOC lev0:
    pl_ = EbasMasterPL()
    for comp in [x.CO_COMP_NAME for x in pl_.META['VOC']]:
        for ext in ('_peak_area', '_peak_width', '_retention_time'):
            compmod = comp + ext
            CO_EXCEPTIONAL['0'][compmod] = {
                'CO_COMP_NAME': compmod,
                'CO_CAPTION': None,
                'CO_DESC': None,
            }
    # PTR-MS lev0:
    for comp in PTR_MS_COMP:
        reg = re.match(r'mass_(.*)(_organic_compounds)?_count_rate', comp)
        if reg:
            capt = 'mz' + reg.group(1)
        else:
            reg = re.match(r'^(.*)_count_rate', comp)
            if reg:
                capt = COOfflineMasterData.META[reg.group(1)].CO_CAPTION
            else:
                RuntimeError(comp + ' missing caption')
        CO_EXCEPTIONAL['0'][comp] = {
            'CO_COMP_NAME': comp,
            'CO_CAPTION': capt,
            'CO_DESC': None,
        }

    def __init__(self, dbh=None):
        """
        Read co masterdata if dbh is provided.
        """
        COOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)


    def __getitem__(self, key):
        """
        Allows dictionary like access to metadata.
        Exception for CO:
            - exceptional metadata lookup (non DB, data level dependent)
        Parameters:
            key    component (string) or tuple (component, data_level)
                   Only component is used for regulat lookup!
                   data_level is optional and is used for finding exceptional
                   data level dependent masterdata
        """
        if isinstance(key, (tuple, list)):
            component = key[0]
            data_level = key[1]
            if data_level in ['0a', '0b']:
                data_level = '0'
        else:
            component = key
            data_level = None
        # first try: use regular masterdata, ignore data level
        try:
            return self.__class__.META[component]
        except KeyError:
            if not data_level:
                raise
        # don't exist: try exceptional masterdata (needs data level)
        # Those masterdata are NOT defined in the database, but might be
        # used e.g. in lev 0 files.
        # Thus they are accepted when reading the file, but the domain
        # layer will issue an error message.
        return self.__class__.CO_EXCEPTIONAL[data_level][component]



"""
ebas/input_output/ebas_csv_write.py
$Id: write.py 2519 2020-10-30 16:13:09Z pe $

output functionality for EBAS CSV module

History:
V.1.0.0  2014-08-07  pe  initial version

"""

import sys
import os
from .base import EbasCSVBase
from ..base import FLAGS_ONE_OR_ALL, FLAGS_NONE
from ebas.domain.basic_domain_logic.time_period import estimate_period_code
from fileformats.CsvUnicode import CsvUnicodeWriter
from ...fileset.xmlwrap import xmlwrap_fileheader, xmlwrap_filetrailer
import textwrap

class EbasCSVPartialWrite(EbasCSVBase):  # pylint: disable=R0901
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    """
    CSV I/O object.

    This is one part of the partial class EbasCSV (output functionality).
    """
    
    def write(self, createfiles=False, destdir=None,  # pylint: disable=R0913
              xmlwrap=False, fileobj=None, flags=FLAGS_ONE_OR_ALL,
              datadef='EBAS_1.1', metadata_options=0, suppress=0):
        # R0913 Too many arguments
        """
        Writes the I/O object to a CSV file (or stdout)
        Parameters:
            createfiles   write to file? else write to stdout
                          (special case for fileobj)
            destdir       destination directory path for files
            xmlwrap       wrap output in xml container
            fileobj       write to filehandle
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
            datadef       ebas data definition (EBAS_1 or EBAS_1.1)
            metadata_options
                          options for output, bitfield:
                          currently only EBAS_IOMETADATA_OPTION_SETKEY
            suppress      suppress selected consolidations (bitfield):
                              SUPPRESS_SORT_VARIABLES
                              SUPPRESS_METADATA_OCCURRENCE
        Returns:
            None
        """
        fil = self.prepare_write(
            True, createfiles=createfiles, destdir=destdir, xmlwrap=xmlwrap,
            fileobj=fileobj, flags=flags, datadef=datadef,
            metadata_options=metadata_options, suppress=suppress)
        self.csv = CsvUnicodeWriter(fil, dialect="excel", encoding="utf-8")
        self.set_header(flags=flags)
        self.set_var_metadata()
        self.__write(flags=flags)
        if xmlwrap:
            xmlwrap_filetrailer()
        if createfiles:
            fil.close()
        self.write_indexdb()

    def set_header(self, flags=FLAGS_ONE_OR_ALL):
        """
        Gnereate the header block of the CSV file.
        Sets directly the CSV attributes.
        Parameters:
            None
        Returns
            None
        """
        self.header = []
        for meta in self.internal.ebasmetadata.metadata_list_tag_valuestr(
                self):
            self.header.append([meta[0], meta[1]])
        if flags == FLAGS_NONE:
            self.csv.writerow([
                'Technical note:',
                textwrap.dedent("""\
            !!! WARNING !!!
            This file has been extracted from EBAS without flag information.
            Data without flag information are easier to process by non expert
            users, but part of the information is lost!
            Changes on extract without flag information:
                - values flagged with at least one flag from the class of
                  invalid flags are changed to MISSING VALUE
                - values flagged with flag 781 ("Value below detection limit,
                  data element contains detection limit") are divided by 2""")])

    def set_var_metadata(self):
        """
        Generates a vname string for the measuremnt column and one for a
        possible flag column for variable vnum.
        Parameters:
            None
        Returns:
            tuple (vname_str, vname_str_flag)
        """
        # cache global metadata
        global_metadata = {
            meta[0]: meta[1]
            for meta in self.internal.ebasmetadata.metadata_list_tag_valuestr(
                self)}

        self.var_metadata = {
            'Variable type': ['Start Time UTC', 'End Time UTC']
        }
        self.var_meta_tags = set()
        self.var_meta_tags.add((-1, 'Variable type'))
        self.var_char_keys = set()
        self.var_characteristics = {}

        for vnum in range(len(self.variables)):
            self.var_metadata['Variable type'].append('Data')
            if 'characteristics' in self.variables[vnum].metadata:
                tags_added = []
                for char in self.variables[vnum].metadata.\
                                characteristics.sorted_order():
                    self.var_char_keys.add((char[0], char[1].CT_TYPE))
                    # tupel (sort_order, ct_type)
                    if char[1].CT_TYPE not in self.var_characteristics:
                        # new characteristic, fill previous vars with ''
                        # start- end time:
                        self.var_characteristics[char[1].CT_TYPE] = ['', '']
                        # other variables:
                        for vnum2 in range(vnum):
                            self.var_characteristics[char[1].CT_TYPE].append('')
                            if self.variables[vnum2].flagcol:
                                self.var_characteristics[char[1].CT_TYPE].\
                                    append('')
                    self.var_characteristics[char[1].CT_TYPE].append(
                        char[1].value_string())
                    tags_added.append(char[1].CT_TYPE)
                for tag in self.var_characteristics.keys():
                    if tag not in tags_added:
                        self.var_characteristics[tag].append('')
            tags_added = []
            for meta in self.internal.ebasmetadata.metadata_list_tag_valuestr(\
                    self, vnum):
                self.var_meta_tags.add((meta[2]['sortorder'], meta[0]))
                # tupel (sort_order, metadata-tag)
                if meta[0] not in self.var_metadata:
                    # new metadata element, fill values for all previous vars
                    self.var_metadata[meta[0]] = ['', '']  # start- and end time
                    # other vars:
                    for vnum2 in range(vnum):
                        if meta[0] in global_metadata:
                            self.var_metadata[meta[0]].append(
                                global_metadata[meta[0]])
                        else:
                            self.var_metadata[meta[0]].append('')
                        if self.variables[vnum2].flagcol:
                            if vnum2 >= 1 and \
                               not self.variables[vnum2-1].flagcol:
                                # make a general flag column
                                self.var_metadata[meta[0]].append('')
                            else:
                                # make a dedicated flag column
                                if meta[0] == 'Unit':
                                    self.var_metadata[meta[0]].append('no unit')
                                else:
                                    self.var_metadata[meta[0]].append(
                                        self.var_metadata[meta[0]][-1])

                self.var_metadata[meta[0]].append(meta[1])
                tags_added.append(meta[0])
            for tag in self.var_metadata.keys():
                if tag != 'Variable type' and tag not in tags_added:
                    if tag in global_metadata:
                        self.var_metadata[tag].append(global_metadata[tag])
                    else:
                        self.var_metadata[tag].append('')
            if self.variables[vnum].flagcol:
                self.var_metadata['Variable type'].append('Flag')
                for tag in self.var_metadata:
                    if tag != 'Variable type':
                        if vnum >= 1 and not self.variables[vnum-1].flagcol:
                            # make a general flag column
                            self.var_metadata[tag].append('')
                        else:
                            # make a dedicated flag column
                            if tag == 'Unit':
                                self.var_metadata[tag].append('no unit')
                            else:
                                self.var_metadata[tag].append(
                                    self.var_metadata[tag][-1])
                for tag in self.var_characteristics:
                    if vnum >= 1 and not self.variables[vnum-1].flagcol:
                        # make a general flag column
                        self.var_characteristics[tag].append('')
                    else:
                        # make a specific flag column (with all characteristics
                        # and metadata)
                        if 'characteristics' in \
                               self.variables[vnum].metadata and \
                           self.variables[vnum].metadata.characteristics \
                               is not None and \
                           self.variables[vnum].metadata.characteristics \
                               .dc_by_ct_type(tag) is not None:
                            self.var_characteristics[tag].append(
                                self.variables[vnum].metadata.characteristics.
                                dc_by_ct_type(tag).value_string())
                        else:
                            self.var_characteristics[tag].append('')


    def __write(self, flags=FLAGS_ONE_OR_ALL):
        """
        Sets the CSV data variables.
        Parameter:
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
        Returns:
            None
        """
        # remove from global header what's included as variable metadata:
        for i in reversed(range(len(self.header))):
            if self.header[i][0] in self.var_metadata:
                del self.header[i]
            else:
                self.header[i][0] += ':'
        # write header
        self.csv.writerow(['EBAS CSV',
                           3+len(self.header)+len(self.var_metadata)+
                           len(self.var_characteristics)+
                           len(self.sample_times)])
        self.csv.writerow(['GLOBAL METADATA', len(self.header)])
        for hea in self.header:
            self.csv.writerow(hea)
        self.csv.writerow(['VARIABLE METADATA AND DATA',
                           len(self.var_metadata)+len(self.var_characteristics),
                           len(self.sample_times)])
        # set the metadata
        for meta in sorted(list(self.var_meta_tags)):
            self.csv.writerow([meta[1]+':'] + self.var_metadata[meta[1]])
        # set the characteristics
        for char in sorted(list(self.var_char_keys)):
            self.csv.writerow([char[1]+':'] + self.var_characteristics[char[1]])

        # prepare values and flags
        values = []
        numflags = []
        for vnum in range(len(self.variables)):
            if flags == FLAGS_NONE:
                # clean values (set invalid to missing, set 781 to 1/2 value)
                # is done only for export - so the I/O object remains
                # fully valid!
                values.append(self.values_without_flags(vnum))
            else:
                values.append(self.variables[vnum].values_)
            if self.variables[vnum].flagcol:
                numflags.append(
                    [float('0.' +\
                           ''.join(["{0:03d}".format(flag)
                                    for flag in flags]))
                     for flags in self.variables[vnum].flags])
            else:
                numflags.append('')
        for i in range(len(self.sample_times)):
            row = ['', self.sample_times[i][0].strftime('%Y-%m-%dT%H:%M:%SZ'),
                   self.sample_times[i][1].strftime('%Y-%m-%dT%H:%M:%SZ')]
            for vnum in range(len(self.variables)):
                row.append(values[vnum][i])
                if self.variables[vnum].flagcol:
                    row.append(numflags[vnum][i])
            self.csv.writerow(row)



"""
ebas/io/csv/base.py
$Id: base.py 2519 2020-10-30 16:13:09Z pe $

EBAS CSV base class

"""

from ..basefile.file import EbasFile
from ..base import EBAS_IOFORMAT_CSV

class EbasCSVBase(EbasFile):  # pylint: disable=R0901
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    """
    Base class for all EBAS CSV classes.
    """

    IOFORMAT = EBAS_IOFORMAT_CSV

    def __init__(self, *args, **kwargs):
        """
        Class initioalization.
        """
        EbasFile.__init__(self, *args, **kwargs)
        self.strict_global = False   # EBAS CSV does NOT use strict_global
        # add additional attributes:
        self.csv = None
        self.header = None
        self.var_metadata = None
        self.var_meta_tags = None
        self.var_characteristics = None
        self.var_char_keys = None

    def gen_filename(self, createfiles=False, destdir=None):
        # pylint: disable-msg=W0221
        # W0221: Arguments number differs from overridden method
        """
        Generates a file name according to NILU file name convention.
        Use the base class functionality but set the file name extension.
        Parameter:
            createfiles   write to physical file?
            destdir       destination directory path for files
        Return:
            None
        """
        EbasFile.gen_filename(self, createfiles=createfiles, destdir=destdir,
                              extension=".csv")

    def _read(self, *args, **kwargs):  # pylint: disable=W0613
        # W0613: Unused argument
        """
        Read is not implemented for csv.
        """
        raise RuntimeError('Read is not implemented for csv.')



"""
$Id: eanet_precip.py 2749 2021-11-25 00:34:17Z pe $

Basic EANET Precipitation functionality is implemented in the EanetPrecip class
in fileformats/EANET.
The ebas/io/eanet_precip (this) module is about mapping the EANET files to the
ebas IO data model.
"""

from .read_precip import EanetPrecipPartialRead

class EbasEanetPrecip(EanetPrecipPartialRead):
    """
    EANET Precip I/O object.
    """




"""
ebas/io/base.py
$Id: base.py 2774 2022-01-10 19:03:29Z pe $

Base classes for (file) I/O.

"""

import datetime
import logging
import os.path
from six import PY3
from copy import deepcopy
from ebas.domain.basic_domain_logic.dataset_types import \
    is_precip_concentration, is_precip_amount, is_auxiliary
from ebas.domain.basic_domain_logic.flags import get_flag_summary
from ebas.io.ebasmetadata import DatasetCharacteristicList
from ebas.domain.basic_domain_logic.unit_convert import UnitConvert, \
    NoConversion
from ebas.domain.basic_domain_logic.time_period import estimate_period_code, \
    estimate_resolution_code
from ebas.domain.basic_domain_logic.parameters import title
from nilutility.datatypes import DataObject
from nilutility.datetime_helper import DatetimeInterval
from nilutility.msg_condenser import MessageRecord, MessageCondenser
from ...ebasmetadata import EbasMetadata
from ebas.io.templates import get_template, NoTemplate

if PY3:
    unicode = str


class EbasMetadataInvalid(unicode):
    """
    Modified unicode class for EBAS metadata. The class is able to contain an
    unicode value in addition to the fact that the value is invalid (by class).
    A metadata parser or check function can keep the original value but set it
    to type EbasMetadataInvalid, thus downstream code will know the value is
    invalid.
    """
    pass

def isempty(var):
    """
    Check if the metadata element is empty in any way.
    Possible types of "empty":
        None    The metadata are not specified in the file or are reported
                as empty string in the file.
        EbasMetadataEmpty
                Is set when the metadata where originally set in the file, but
                the parser experienced an error. This special value is set to
                avoid false errors about mandatory metadata missing.
    """
    if var is None or isinstance(var, EbasMetadataInvalid):
        return True
    return False

def isvalid(var):
    """
    Check if the metadata element contains a valid value.
    Opposite of isempty.
    """
    return not isempty(var)

class EbasIOFileMessageRecord(MessageRecord):
    """
    Derived MessageRecord class for EbasIO:
    Adds column number to keys and line number to data.
    """
    KEYS = ['msgfunc', 'msg_id', 'vnum']
    DATA = ['message', 'line']


class EbasIOFileMessageCondenser(MessageCondenser):
    """
    Message condenser for NasaAmes logging
    """

    MSG_ID_FLAG_MISSING = 1
    MSG_ID_FLAG_ZERODOT = 2
    MSG_ID_TIME_OVERLAP = 3
    MSG_ID_TIME_MISSING = 4
    MSG_ID_TIME_INVERT = 5

    def deliver_msg(self, record, condensed):
        """
        Deliver a single message.
        Parameters:
            record     the message record
            condensed  bool, condensed message
        Returns:
            None
        """
        record.msgfunc(record.message, lnum=record.line)
        if condensed:
            lst = self.dict[record.key]
            record.msgfunc(
                "The previous message repeats {} times{}. First occurrences "
                "in lines: {}, ..."
                .format(len(lst)-1,
                        " for variable {}".format(record.vnum) \
                            if record.vnum is not None else "",
                        ", ".join(
                            [str(x.line) for x in lst[0:self.threshold]])),
                lnum=record.line, counter_increase=len(lst)-1)
            # use same lnum as above message, otherwise the lines get parted
            # in submit-tool


class EbasFileBase(object): # pylint: disable-msg=R0902
    # R0902: Too many instance attributes
    """
    Base class for all EBAS I/O objects.
    """

    IOFORMAT = None  # Must be set by derived classes

    def __init__(self, indexdb=None, condense_messages=0):
        """
        Constructor for class IOBase.
        Parameters:
            indexdb           index db object (if indexdb should be written)
            condense_messages threshold for condensing repeating messages.
                              0 turns off condensation.
        """
        self.logger = logging.getLogger(self.__class__.__name__)

        # strict global metadata: this is considered part of the file format
        # specification and must be set in the __init__ of all derived classes.
        # (only metadata which apply to ALL variables are considered global)
        self.strict_global = None

        self.indexdb = indexdb
        self.msg_condenser = EbasIOFileMessageCondenser(
            threshold=condense_messages, record_cls=EbasIOFileMessageRecord)
        # TODO: most of the internal attributes are only used by one of the
        # derived classes (NasaAmes parser, fromdomain, todomain etc)
        self.from_domain = False
        self.internal = \
            DataObject(
                # metadata.timeref as timedelta object
                timeoffset=datetime.timedelta(seconds=0),
                # temporary attributes used during reading in different
                #implementations:
                read_tmp=None,
                # temporary attributes used in from_domain:
                fromdomain_tmp=None,
                # used internaly for I/O of ebas specific metadata,
                # will be initialized to an EbasMetadata object on read/write
                ebasmetadata=None,
                )
        self.metadata = \
            DataObject(setkey=None,    # dataset id (needed for xml export)
                       datadef=None,   # specification of file format
                       license=None,   # license URI
                       citation=None,  # citation string
                       filename=None,  # generated filename
                       doi=None,       # doi URI and optional part number
                       # [doi, part]
                       # (part number of the doi) if DOI is exported to multiple
                       # files
                       doi_list=None,  # list of DOIs (series state) the file
                       # contains data from []
                       creation_time=None, # file creation date (datetime)
                       export_state=None,  # export state for DB extracts
                       export_filter=None,  # export filter (f900, invalid)
                       startdate=None, # first start time (datetime)
                       timezone=None,  # timezone, currently only UTC
                       org=None,       # ORG code (string)
                       originator=[],  # list of EbasDB PS dict.
                       submitter=[],   # list of EbasDB PS dict.
                       projects=[],    # list of PR_ACRONYMS (strings)
                       # list of project names, descriptions, contact names and
                       # contact emails: only for output and for non-NasaAmes:
                       project_names = [],
                       project_descs = [],
                       project_cotact_names = [],
                       project_contact_emails = [],
                       type=None,      # 'TU' / 'TI' (string)
                       revision=None,  # revision number (string)
                       revdate=None,   # latest update data or meta (datetime)
                       revdesc=None,   # revision desc (string)
                       input_dataset=None, # list of strings
                       software=None,  # software (string)
                       acknowledgements=None, # (string)
                       reference_date=None, # zero point of time axes (datetime)
                       timeref=None,   # timeref HH_MM (string)
                       period=None,
                       resolution=None,
                       rescode_sample=None,
                       duration=None,
                       datalevel=None,
                       regime=None,
                       matrix=None,
                       comp_name=None, # component name
                       unit=None,
                       statistics=None,# statistics code
                       characteristics=None, # ds characteristics (dict)
                       title=None,     # title name (comp.caption or comp.name)
                       station_code=None,
                       platform_code=None, # (7 chars including 6 char st.code)
                       station_name=None,
                       station_wdca_id=None,
                       station_gaw_name=None,
                       station_gaw_id=None,
                       station_airs_id=None,
                       station_other_ids=None,
                       station_state_code=None,
                       station_landuse=None,
                       station_setting=None,
                       station_gaw_type=None,
                       station_wmo_region=None,
                       station_latitude=None,
                       station_longitude=None,
                       station_altitude=None,
                       mea_latitude=None,
                       mea_longitude=None,
                       mea_altitude=None,
                       mea_height=None,
                       instr_type=None,
                       instr_pid=None,
                       lab_code=None,
                       instr_name=None,
                       instr_manufacturer=None,
                       instr_model=None,
                       instr_serialno=None,
                       sensor_type=None,
                       ana_technique=None,
                       ana_lab_code=None,
                       ana_instr_name=None,
                       ana_instr_manufacturer=None,
                       ana_instr_model=None,
                       ana_instr_serialno=None,
                       cal_scale=None,
                       cal_std_id=None,
                       sec_std_id=None,
                       inlet_type=None,
                       inlet_desc=None,
                       inlet_tube_material=None,  # string, restricted
                       inlet_tube_outerD=None,    # float, unit=mm
                       inlet_tube_innerD=None,    # float, unit=mm
                       inlet_tube_length=None,    # float, unit=m
                       maintenance_desc=None,     # string, freetext
                       hum_temp_ctrl=None,
                       hum_temp_ctrl_desc=None,
                       vol_std_temp=None,
                       vol_std_pressure=None,
                       detection_limit=None,  # (value, unit)
                       detection_limit_desc=None,
                       upper_range_limit=None,  # (value, unit)
                       uncertainty=None,  # (value, unit)
                       uncertainty_desc=None,
                       zero_negative=None,
                       zero_negative_desc=None,
                       ext_lab=None,
                       method=None,
                       std_method=None,
                       medium=None,
                       flow_rate=None,
                       filter_face_velocity=None,
                       filter_area=None,
                       filter_descr=None,
                       filter_prefiring=None,
                       filter_conditioning=None,
                       coating_solution=None,
                       sample_prep=None,
                       blank_corr=None,
                       artifact_corr=None,
                       artifact_corr_desc=None,
                       charring_corr=None,
                       ozone_corr=None,
                       watervapor_corr=None,
                       zero_span_type=None,       # string: automatic/manual
                       zero_span_interval=None,   # period code, e.g 1d
                       abs_cross_section=None,    # float (unit=cm2)
                       qa=None,                   # dictionary, many QA blocks
                       # example:
                       # [
                       #   DataObject({
                       #    'qa_number': 1,
                       #    'qm_id': '...',
                       #    'qa_date': datetime.datetime(...),
                       #    ...}),
                       #   DataObject(...)
                       #   ...
                       # ]
                       comment=None,

                       # elements only supporded in lev 0 (implemented in
                       # ebasmetadata with condition data_level == 0
                       # those elements are NOT implemented in ebas.domain!
                       time_inlet_to_converter=None,
                       time_converter_or_bypass_line=None,
                       time_stay_converter= None,
                       converter_temp= None,
                       filter_type=None,
                       multi_scattering_corr_fact=None,
                       max_attenuation=None,
                       leakage_factor_zeta=None,
                       comp_thresh_atten1=None,
                       comp_thresh_atten2=None,
                       comp_param_kmin=None,
                       comp_param_kmax=None,
                       mass_abs_cross_section=None,
                       coincidence_corr = None,
                       charge_type = None,
                       inlet_diffusion_loss_data = None,
                       cpc_default_pulse_width = None,
                       normalized_ion_mobility = None,
                       number_density_of_air = None,
                       normalized_pressure = None,
                       normalized_count_rate = None,
                       drift_tube_length = None,
                      )

        self.sample_times = []
        # self.sample times: list of tupels (start, end)
        # (subsetted to correct time interval)

        self.variables = []
        # self.variables: list of DataObject:
        #  {
        #    is_hex:    True for hexint variables (e.g. instrument status)
        #    values_:   measurement values (list)
        #               (subsetted to correct time interval)
        #    flags:     flags for measurments (list of tuples)
        #               (subsetted to correct time interval)
        #    flagcol:   Bool, whether this variable needs a flag column
        #    flagtitle: Text, used as title for flag column
        #               (analog to metadata.title for the variable itself)
        #    metadata   variable specific metadata
        #               (potentially the same keys as self.metadata, but only if
        #                it differs from self.metadata)
        #
        #    internal metadata (only used on demand):
        #
        #  }

        # attributes needed for error handling
        self.errors = 0
        self.warnings = 0

    def from_ebasfile(self, ebasfile):
        """
        Clone object from another ebas file objects.
        """
        self.metadata = deepcopy(ebasfile.metadata)
        self.sample_times = deepcopy(ebasfile.sample_times)
        self.variables = deepcopy(ebasfile.variables)

    def setup_non_domain(self):
        """
        """
        self.internal.nondomain_tmp = DataObject(id_set=set())

    def merge_var_non_domain(self, id, metadata, sample_times, values, flags,
            convert, is_hex=False):
        """
        For data source other than ebas domain
        Add variable to file if it fits (metadata and the timesseries must be
        matching).
        Parameters:
            metadata        metadata dict (DataObject)
            sample_times    sample times list
            values          values list
            flags           flags list
            convert         unit conversion: 0 none, 1 convert,
                                             2 add additional converted variable
            is_hex          should be represented in hex
        Returns:
            (state, remaining)
            state: True (could be added), False (could not be added)
            remaining:
                None in case of state==Flase
                else: list of remaining intervals, each tuple
                      (saple_times, values, flags)
        """
        if len(sample_times) != len(values) or len(sample_times) != len(flags):
            raise RuntimeError("different sample times and values/flags")
        if not sample_times and self.sample_times or \
                sample_times and not self.sample_times:
            # samples do not overlap
            return False, None
        if sample_times and \
                (sample_times[0][0] > self.sample_times[0][0] or \
                 sample_times[-1][1] < self.sample_times[-1][1]):
            return False, None
        start = 0
        for i in range(len(sample_times)):
            if sample_times[i] == self.sample_times[0]:
                start = i
                break
        else:
            return False, None
        if self.sample_times != \
                sample_times[start:len(self.sample_times)+start]:
            return False, None

        # fix metadata defaults:
        if 'org' not in metadata:
            metadata.org = None
        if 'station_code' not in metadata:
            metadata.station_code = None
        if 'lab_code' not in metadata:
            metadata.lab_code = None
        if 'instr_name' not in metadata:
            metadata.instr_name = None
        if 'matrix' not in metadata:
            metadata.matrix = None
        if 'comp_name' not in metadata:
            metadata.comp_name = None
        if 'originator' not in metadata:
            metadata.originator = []
        if 'projects' not in metadata:
            metadata.projects = []

        # metadata that may NOT differ in one file:
        # organisation, station, originators, submitters, projects,\
        if self.get_meta_for_var(0, 'org') != metadata.org or \
            self.get_meta_for_var(0, 'station_code') != metadata.station_code:
            return False, None
        # generally avoid different instruments. but allow for:
        #  - precip amount
        #  - matrix position
        if self.get_meta_for_var(0, 'lab_code') != metadata.lab_code or \
            self.get_meta_for_var(0, 'instr_name' != metadata.instr_name):
            # this would be a new instrument for this file
            if metadata.matrix != 'position' and \
               not any([self.get_meta_for_var(i, 'matrix') == 'position'
                        for i in range(len(self.variables))]):
                if (not is_precip_amount(metadata.matrix, metadata.comp_name) and \
                    not is_precip_concentration(metadata.matrix, metadata.comp_name)) or \
                   (is_precip_amount(metadata.matrix, metadata.comp_name) and not \
                    any([self.is_precip_concentration(i) for i in range(len(self.variables))])) or \
                   (is_precip_concentration(metadata.matrix, metadata.comp_name) and not \
                    any([self.is_precip_amount(i) for i in range(len(self.variables))])):
                    return False, None

        if set(tuple([tuple(sorted(x.items()))
                      for x in self.get_meta_for_var(0, 'originator')])) != \
           set(tuple([tuple(sorted(x.items()))
                      for x in metadata.originator])):
            return False, None
        # submission metadata can go to the same nasa_ames file
        # Set of projects must be checked in any cas. Can be changed afterwards,
        # thus the same submission is _not_ sufficient.
        if set(self.get_meta_for_var(0, 'projects')) != set(metadata.projects):
            return False, None

        # if this variable is not already in the file, add the variable:
        if id not in self.internal.nondomain_tmp.id_set:
            self._add_variable_nondomain(
                id,
                metadata, 
                values[start:start+len(self.sample_times)],
                flags[start:start+len(self.sample_times)],
                convert,
                is_hex=is_hex)
        # calculate remaining interval(s):
        remain = []
        if  start > 0:
            remain.append((sample_times[0:start], values[0:start],
                flags[0:start]))
        if len(sample_times) > start + len(self.sample_times):
            rest = start + len(self.sample_times)
            remain.append((sample_times[rest:], values[rest:], flags[rest:]))
        return True, remain

    def _add_variable_nondomain(self, id, metadata, values, flags, convert,
        is_hex=False):
        """
        Add variable to file (other sources than ebas domain).
        Parameters:
            metadata        metadata dict (DataObject)
            values          values list
            flags           flags list
            convert         unit conversion: 0 none, 1 convert,
                                             2 add additional converted variable
            is_hex          should be represented in hex
        Returns:
            None
        """
        self.internal.nondomain_tmp.id_set.add(id)
        self.variables.append(
            DataObject(
                values_=values, flags=flags, flagcol=True,
                metadata=DataObject(metadata), is_hex=is_hex))
        if convert == 1:
            # convert variable if needed:
            self._convert_variable_output_new()
        if convert == 2:
            try:
                vol_std_temp = float(metadata.vol_std_temp)
            except (ValueError, TypeError, AttributeError):
                # ValueError if string, eg instrument internal
                # TypeError if None
                # AttributeError if not in metadsta
                vol_std_temp = None
            try:
                vol_std_pressure = float(metadata.vol_std_pressure)
            except (ValueError, TypeError, AttributeError):
                # ValueError if string, eg instrument internal
                # TypeError if None
                # AttributeError if not in metadsta
                vol_std_pressure = None
            regime = metadata.regime if 'regime' in metadata else None
            matrix = metadata.matrix if 'matrix' in metadata else None
            comp_name = metadata.comp_name if 'comp_name' in metadata else None
            unit = metadata.unit if 'unit' in metadata else None
            try:
                unit_convert = UnitConvert()
                unit_convert.export_conv_params(
                    regime, matrix, comp_name, unit,
                    vol_std_pressure, vol_std_temp)
                # We do not care about the result here. We just know there is a
                # possible conversion if there is no NoConversion exception.
            except NoConversion as expt:
                for msg in expt.msgs:
                    self.logger.warning(
                        "variable {}: {}".format(
                            len(self.variables) + 1, msg))
            else:
                # add converted variable if needed and conversion is possible
                # make copies of value and flag lists, otherwise the values
                # of the original variable gets converted as well
                self._add_variable_nondomain(
                    id, metadata, list(values), [[x for x in f] for f in flags],
                    1, is_hex=is_hex)

    def _convert_variable_output_new(self):
        """
        Adds one variable to file if conversion is needed.
        This is a new experimental implementation for files without ebas domain.
        This can in principle also be used for domain export.
        TODO: retire the old _convert_variable_output method
        Parameters:
            None
        Returns:
            None
        """
        regime = self.get_meta_for_var(-1, 'regime')
        matrix = self.get_meta_for_var(-1, 'matrix')
        comp_name = self.get_meta_for_var(-1, 'comp_name')
        unit = self.get_meta_for_var(-1, 'unit')
        try:
            vol_std_temp = float(self.get_meta_for_var(-1, 'vol_std_temp'))
        except (ValueError, TypeError):
            vol_std_temp = None
        try:
            vol_std_pressure = float(
                self.get_meta_for_var(-1, 'vol_std_pressure'))
        except (ValueError, TypeError):
            vol_std_pressure = None
        try:
            unit_convert = UnitConvert()
            (conv_obj, reset_stdcond) = unit_convert.export_conv_params(
                regime, matrix, comp_name, unit, vol_std_pressure, vol_std_temp)
        except NoConversion as expt:
            for msg in expt.msgs:
                self.logger.warning(
                    "variable {}: {}".format(
                        len(self.variables) + 1, msg))
            return
        if self.__class__.LAZY_READ:
            # fake the conversion (it should be delayed in the lazy read object)
            # TODO: we use hardcoded 4 decimals for now, are there better ways?
            conv_obj.convert_data([1.0001, 1.0002])
            self.variables[-1].values_.conversion_functions.append(
                conv_obj.convert_data)
        else:
            # convert the values right away:
            conv_obj.convert_data(self.variables[-1].values_)
        cnvstr = conv_obj.conversion_string()
        self.logger.info(
            "variable {}: Componment '{}', converting unit {}.".format(
                len(self.variables) + 1, comp_name, cnvstr))
        conv_remark = "Data converted on export from EBAS {}.".format(
            cnvstr)

        # correct uncertainty:
        uncert = self.get_meta_for_var(-1, 'uncertainty')
        if uncert and uncert[1] == unit:
            origval = uncert[0]
            if conv_obj.rounding is None:
                # rounding needs to be reset after conversion
                reset = True
            else:
                reset = False
            value_list = [origval]
            conv_obj.convert_data(value_list)
            newval = value_list[0]
            self.variables[-1].metadata.uncertainty = (newval, conv_obj.to_unit)
            cnvstr = conv_obj.conversion_string(from_val=origval,
                                                to_val=newval)
            self.logger.info(
                "variable {}: converting 'Measurement uncertainty' for "
                "component {} {}.".format(len(self.variables) + 1,
                                          comp_name, cnvstr))
            conv_remark += " Measurement uncertainty converted {}.".format(
                cnvstr)
            if reset:
                conv_obj.rounding = None  # reset the rounding

        # correct detection limit:
        detect = self.get_meta_for_var(-1, 'detection_limit')
        if detect and detect[1] == unit:
            origval = detect[0]
            if conv_obj.rounding is None:
                # rounding needs to be reset after conversion
                reset = True
            else:
                reset = False
            value_list = [origval]
            conv_obj.convert_data(value_list)
            newval = value_list[0]
            self.variables[-1].metadata.detection_limit = (newval,
                                                           conv_obj.to_unit)
            cnvstr = conv_obj.conversion_string(from_val=origval, to_val=newval)
            self.logger.info(
                "variable {}: converting 'Detection limit' for component {} {}."
                .format(len(self.variables) + 1, comp_name,
                        cnvstr))
            conv_remark += " Detection limit converted {}.".format(cnvstr)
            if reset:
                conv_obj.rounding = None  # reset the rounding

        # correct upper range limit:
        upper = self.get_meta_for_var(-1, 'upper_range_limit')
        if upper and upper[1] == unit:
            origval = upper[0]
            if conv_obj.rounding is None:
                # rounding needs to be reset after conversion
                reset = True
            else:
                reset = False
            value_list = [origval]
            conv_obj.convert_data(value_list)
            newval = value_list[0]
            self.variables[-1].metadata.upper_range_limit = (newval,
                                                             conv_obj.to_unit)
            cnvstr = conv_obj.conversion_string(from_val=origval, to_val=newval)
            self.logger.info(
                "variable {}: converting 'Upper range limit' for component "
                "{} {}." .format(len(self.variables) + 1,
                                 comp_name, cnvstr))
            conv_remark += " Upper range limit {}.".format(cnvstr)
            if reset:
                conv_obj.rounding = None  # reset the rounding

        # correct QA bias and variability:
        for qanum in self.get_qa_numbers():
            qa_ = self.get_qa_by_number(qanum, -1)
            if 'qa_bias' in qa_ and qa_.qa_bias and qa_.qa_bias[1] == unit:
                if conv_obj.rounding is None:
                    # rounding needs to be reset after conversion
                    reset = True
                else:
                    reset = False
                origval = qa_.qa_bias[0]
                value_list = [origval]
                conv_obj.convert_data(value_list)
                newval = value_list[0]
                qa_.qa_bias = (newval, conv_obj.to_unit)
                cnvstr = conv_obj.conversion_string(from_val=origval,
                                                    to_val=newval)
                self.logger.info(
                    "variable {}: converting 'QA{} bias' for component "
                    "{} {}." .format(
                        len(self.variables) + 1,
                        str(qanum) if qanum == -1 else '',
                        comp_name, cnvstr))
                conv_remark += " QA{} bias {}.".format(
                    str(qanum) if qanum == -1 else '', cnvstr)
                if reset:
                    conv_obj.rounding = None  # reset the rounding
            if 'qa_variability' in qa_ and qa_.qa_variability and \
                    qa_.qa_variability[1] == unit:
                if conv_obj.rounding is None:
                    # rounding needs to be reset after conversion
                    reset = True
                else:
                    reset = False
                origval = qa_.qa_variability[0]
                value_list = [origval]
                conv_obj.convert_data(value_list)
                newval = value_list[0]
                qa_.qa_variability = (newval, conv_obj.to_unit)
                cnvstr = conv_obj.conversion_string(from_val=origval,
                                                    to_val=newval)
                self.logger.info(
                    "variable {}: converting 'QA{} variability' for component "
                    "{} {}." .format(
                        len(self.variables) + 1,
                        str(qanum) if qanum == -1 else '',
                        comp_name, cnvstr))
                conv_remark += " QA{} variability {}.".format(
                    str(qanum) if qanum == -1 else '', cnvstr)
                if reset:
                    conv_obj.rounding = None  # reset the rounding

        # correct metadata:
        self.variables[-1].metadata.unit = conv_obj.to_unit
        if reset_stdcond:
            self.variables[-1].metadata.vol_std_pressure = reset_stdcond[0]
            self.variables[-1].metadata.vol_std_temp = reset_stdcond[1]
        oldcomment = self.get_meta_for_var(-1, 'comment')
        if oldcomment:
            conv_remark = oldcomment + " -- " + conv_remark
        self.variables[-1].metadata.comment = conv_remark

    def error(self, msg, lnum=None, counter_increase=1, prefix=''):
        """
        Error handling: Error will be logged, counter incremented.
        Parameters:
            msg        error message
            lnum       line number, optional (only if file is related to a
                       physical file (read))
            counter_incrrease
                       increase the file objects error counter by n
                       (default 1, other values used by message condenser)
            prefix     prefix message, comes before line number
        Returns:
            None
        """
        self.errors += counter_increase
        if lnum is not None:
            msg = u"line {}: {}".format(lnum, msg)
        self.logger.error(prefix + msg)

    def warning(self, msg, lnum=None, counter_increase=1, prefix=''):
        """
        Warning will be logged, counter incremented.
        Parameters:
            msg         warning message
            lnum        line number, optional (only if file is related to a
                        physical file (read))
            counter_incrrease
                       increase the file objects error counter by n
                       (default 1, other values used by message condenser)
            prefix     prefix message, comes before line number
        Returns:
            None
        """
        self.warnings += counter_increase
        if lnum is not None:
            msg = u"{}line {}: {}".format(prefix, lnum, msg)
        self.logger.warning(prefix + msg)

    def gen_filename(self, createfiles=False, destdir=None, extension=""):
        """
        Generates a file name according to NILU file name convention.
        The file name is created according to objects metadata, and the
        resulting filename is also set as metadata attribute.

        If parameter createfiles is set to True, the filename is also tested for
        uniqueness. If needed, _dupN is added (where N is an ascending integer)

        Parameter:
            createfiles   write to physical file?
            destdir       destination directory path for files
            extension     file name extension
        Return:
            None
        """
        def _add(elem):
            if isvalid(elem):
                return elem
            return ''

        if isvalid(self.metadata.doi) and self.metadata.doi[1]:
                # DOI part number is set
                # use doi + part number, fall back to dup if exists
                from ebas.domain.entities.do import DOI_PREFIX
                # import here, to make ebas-io independent of domain
                fname = '{}_part{:02.0f}'.format(
                    self.metadata.doi[0].replace(DOI_PREFIX, ''),
                    self.metadata.doi[1])
        elif isvalid(self.metadata.doi) and self.metadata.doi[0]:
                # use the doi as file name, fallback to _dup if it exists
                from ebas.domain.entities.do import DOI_PREFIX
                # import here, to make ebas-io independent of domain
                fname = self.metadata.doi[0].replace(DOI_PREFIX, '')
        else:
            fi_ref = ''
            lev = ''
            startdate = ''
            revdate = ''
            if isvalid(self.metadata.lab_code) and \
               isvalid(self.metadata.instr_name):
                fi_ref = self.metadata.lab_code + '_' + self.metadata.instr_name
                # TODO: temporary fix, to avoid '/' in filenames.
                # Needs final fix: - change existing FI_REF with /
                #                  - change DB constraint
                #                  - change input checks
                fi_ref = fi_ref.replace('/', '-')
            if isvalid(self.metadata.datalevel):
                lev = 'lev' + self.metadata.datalevel
            if isvalid(self.metadata.startdate):
                startdate = self.metadata.startdate.strftime("%Y%m%d%H%M%S")
            if isvalid(self.metadata.revdate):
                revdate = self.metadata.revdate.strftime("%Y%m%d%H%M%S")
    
            fname = _add(self.metadata.station_code) + '.' +\
                startdate + '.' +\
                revdate + '.' +\
                _add(self.metadata.instr_type) + '.' +\
                _add(self.metadata.comp_name) + '.' +\
                _add(self.metadata.matrix) + '.' +\
                _add(self.metadata.period) + '.' + \
                _add(self.metadata.resolution) + '.' + \
                fi_ref + '.' +\
                _add(self.metadata.method) + '.' +\
                lev
        testname = fname + extension
        if createfiles:
            if destdir is None:
                destdir = '.'
            i = 1
            while os.path.exists(os.path.join(destdir, testname)):
                testname = fname + '_dup{0}'.format(i) + extension
                i += 1
        self.metadata.filename = testname

    def setup_ebasmetadata(self, datadef, metadata_options):
        """
        Set up the ebasmetadata structure depending on format, datadef, and
        metadata options.
        Parameters:
            datadef    data definition (EBAS_1 or EBAS_1.1)
            metadata_options
                       options for output (bitfield):
                       EBAS_IOMETADATA_OPTION_SETKEY
        Returns
            None
        """
        # must be implemented here, because ebasmetadata can't be imported in
        # ..base
        self.metadata.datadef = datadef
        self.internal.ebasmetadata = EbasMetadata(
            self.metadata.datadef, self.__class__.IOFORMAT,
            data_level=self.metadata.datalevel,
            metadata_options=metadata_options)

    def set_default_metadata(self):
        """
        Sets some default metadata (non mandatory metadata elements which have
        a defined default.
        Is used both, for import and export.
        Parameters:
            None
        Returns:
            None
        """
        # some metadata must be set strictly for the whole file
        # make sure those are not set in any variable:
        reset_meta = ['timezone', 'creation_time']
        for var in self.variables:
            for meta in reset_meta:
                if meta in var.metadata:
                    del var.metadata[meta]

        if not self.metadata.timezone:
            self.metadata.timezone = 'UTC'
        if not self.metadata.regime:
            self.metadata.regime = 'IMG'
        now = datetime.datetime.utcnow()
        if not self.metadata.creation_time:
            self.metadata.creation_time = now
        if not self.metadata.revdate:
            self.metadata.revdate = now
        if not self.metadata.platform_code and self.metadata.station_code:
            self.metadata.platform_code = self.metadata.station_code[0:6] + 'S'
        if self.metadata.statistics is None:
            self.metadata.statistics = 'arithmetic mean'

    def get_meta_for_var(self, vnum, elem):
        """
        Get the applicable metadata element for a specific variable.
        (First lookup file metadata, then override with var metadata)
        Parameters:
            vnum   variable index
            elem   metadata element needed
        Returns:
            metadata elemnt value
        """
        res = None
        if elem in self.metadata:
            res = self.metadata[elem]
        if elem in self.variables[vnum].metadata:
            res = self.variables[vnum].metadata[elem]
        return res

    def get_characteristics_for_var(self, vnum):
        """
        Get all characteristics for a specific variable.
        (First lookup file metadata, then override with var metadata)
        Parameters:
            vnum   variable index
            elem   metadata element needed
        Returns:
            metadata elemnt value
        """
        ret = {}
        if self.metadata.characteristics:
            for cha in self.metadata.characteristics:
                if cha.CT_DATATYPE == 'CHR':
                    ret[cha.CT_TYPE] = cha.DC_VAL_CHR
                elif cha.CT_DATATYPE == 'DBL':
                    ret[cha.CT_TYPE] = cha.DC_VAL_DBL
                else:
                    ret[cha.CT_TYPE] = cha.DC_VAL_INT
        if 'characteristics' in self.variables[vnum].metadata and \
           self.variables[vnum].metadata.characteristics:
            for cha in self.variables[vnum].metadata.characteristics:
                if cha.CT_DATATYPE == 'CHR':
                    ret[cha.CT_TYPE] = cha.DC_VAL_CHR
                elif cha.CT_DATATYPE == 'DBL':
                    ret[cha.CT_TYPE] = cha.DC_VAL_DBL
                else:
                    ret[cha.CT_TYPE] = cha.DC_VAL_INT
        return ret

    def add_var_characteristics(self, vnum, ct_type, value):
        """
        Add a characteristic for a variable.
        Parameters:
            vnum      variable index
            ct_type   characteristics type (tag)
            value     value for the characteristic
        """
        if 'characteristics' not in self.variables[vnum].metadata:
            self.variables[vnum].metadata.characteristics = \
                DatasetCharacteristicList()
        ft_type = self.get_meta_for_var(vnum, 'instr_type')
        co_comp_name = self.get_meta_for_var(vnum, 'comp_name')
        data_level = self.get_meta_for_var(vnum, 'datalevel')
        self.variables[vnum].metadata.characteristics.add(
            ct_type, value, ft_type, co_comp_name, data_level=data_level)

    def get_metadata_list_per_var(self, element, string=False):
        """
        Generates a list for one metadata element (one list element per variable
        in file).
        Use variable metadata if possible, else global metadata.
        
        !!! EbasJoinedFile: (hidden catch)
        This method works only for metadata elements that are duplicated to the
        joined file (i.e. elements that are considered to be common) in the join
        method
         
        Parameter:
            element     metadata element (string)
        Returns:
            lst (one element per variable in file, None for empty)
        """
        lst = []
        for vnum in range(len(self.variables)):
            if element in self.variables[vnum].metadata:
                litem = self.variables[vnum].metadata[element]
            else:
                litem = self.metadata[element]
            if string:
                litem = str(litem)
            lst.append(litem)
        return lst

    def get_metadata_set(self, element, string=False, exclude_aux=False):
        """
        Generates a set of different values for one metadata element (of each
        variable).
        This method is meant to give "all different" values for a metadata
        element, this behavior is also overridden in the EbasJoinedFile object.
        Parameter:
            element     metadata element (string)
            string      whether the elements should be reresented as str
            exclude_aux
                        do not report metadata for auxiliary variables
        Returns:
            list (list with unique elements, i.e. set)
        """
        # Because metadata values can be unhashable objects (dict, DataObject),
        # this method must be implemented as a list, not as a set.
        set_ = []
        _str = lambda x: str(x) if string else x

        def _add(x):
            if x is not None and _str(x) not in set_: set_.append(_str(x))

        for vnum in range(len(self.variables)):
            if exclude_aux and self.is_auxiliary(vnum):
                continue
            if isinstance(element, tuple):
                # element is an iterator -- list combinations
                res = []
                for atom in element:
                    res.append(self.get_meta_for_var(vnum, atom))
                _add(res)
            else:
                # else, element is a single metadata key
                res = self.get_meta_for_var(vnum, element)
                if res is not None:
                    _add(res)
        return set_

    def is_precip_concentration(self, vnum):
        """
        Check if the dataset is a precipitation concentration component (one
        that needs precipitation amount)
        Parameters:
            vnum    variable number
        Returns:
            True/False
        """
        return is_precip_concentration(self.get_meta_for_var(vnum, 'matrix'),
                                       self.get_meta_for_var(vnum, 'comp_name'))

    def is_precip_amount(self, vnum):
        """
        Check if the dataset is a precipitation amount component
        Parameters:
            vnum    variable number
        Returns:
            True/False
        """
        return is_precip_amount(self.get_meta_for_var(vnum, 'matrix'),
                                self.get_meta_for_var(vnum, 'comp_name'))

    def is_auxiliary(self, vnum):
        """
        Check if the dataset is to be considered "auxiliary data".
        Parameters:
            vnum    variable number
        Returns:
            True/False
        """
        return is_auxiliary(self.get_meta_for_var(vnum, 'matrix'),
                            self.get_meta_for_var(vnum, 'comp_name'))

    def get_submitter_emails(self):
        """
        Get the email addresses of all data submitters.
        Parameters:
            None
        Returns:
            list of email addresses
        """
        ret = set()
        for ps_ in self.metadata.submitter:
            if 'PS_EMAIL' in ps_ and ps_['PS_EMAIL']:
                ret.add(ps_['PS_EMAIL'])
        return list(ret)


    def get_qa_numbers(self):
        """
        Get a all used QA numbers in the file.
        Sorted by order in main metadata, then by order in each variable as
        they appear in the file.
        Parameters:
            None
        Returns:
            List of integers
        """
        ret_list = []
        if 'qa' in self.metadata and self.metadata.qa is not None:
            for qa_ in self.metadata.qa:
                if qa_['qa_number'] not in ret_list:
                    ret_list.append(qa_['qa_number'])
        for var in self.variables:
            if 'qa' in var.metadata and var.metadata.qa is not None:
                for qa_ in var.metadata.qa:
                    if qa_['qa_number'] not in ret_list:
                        ret_list.append(qa_['qa_number'])
        return ret_list

    def get_qa_by_number(self, qa_number, var_index):
        """
        Get the index for the qa element with a specific qa_number.
        Parameters:
            qa number     number of QA block
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            qa element with the matching qa_number
            self._empty_qa() if not present and var_indes is None
            empty DataObject if not present and var_index is not None
        Raises:
            RuntimeError if more then one list element matches the qa_number
        """
        if var_index is None:
            metadata = self.metadata
        else:
            metadata = self.variables[var_index].metadata
        if 'qa' not in metadata or metadata.qa is None:
            return self._empty_qa() if var_index is None else DataObject()
        qa_list = [qa for qa in metadata.qa if qa['qa_number'] == qa_number]
        if len(qa_list) > 1:
            # this should never happen
            raise RuntimeError('QA metadata qa_number/index error')
        if len(qa_list) == 0:
            return self._empty_qa() if var_index is None else DataObject()
        return qa_list[0]
    
    def get_qa_for_var(self, vnum):
        """
        Returns a list with all integrated qa metadata foe one variable.
        (Fileglogal / per varaible settings are merged)
        Parameters:
            vnum    variable number
        Returns:
            generator of the merged qa dictionaries.
            ! Empty qa dictionaries are skipped !
        """
        for qanum in self.get_qa_numbers():
            qa_ = self.get_qa_by_number(qanum, None)
            if qa_:
                qa_ = qa_.copy()  # need to copy because of update below!
            else:
                qa_ = {}
            var_qa = self.get_qa_by_number(qanum, vnum)
            if var_qa:
                qa_.update(var_qa)
            if not self._is_empty_qa(qa_):
                yield qa_

    def _empty_qa(self):
        """
        Generate an empty qa dictionary for global masterdata (all keys, but all
        values are None).
        """
        return DataObject({x["key"]: None
                           for x in self.internal.ebasmetadata.metadata
                           if "QA_block" in x and x["QA_block"]})

    def _is_empty_qa(self, qa_):
        """
        Check if qa dictionary is empty.
        Returns:
           True if empty, else Flase
        """
        return not any([True for key in qa_.keys()
                        if key != 'qa_number' and qa_[key] is not None])

    def set_qa(self, key, value, qa_number, var_index):
        """
        Set a metadata value for a QA metadata element.
        Check if element has been set before (log error if set already)
        Parameters:
            key           metadata key
            value         value to be set
            qa number     number of QA block
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            None
        """
        if var_index is None:
            metadata = self.metadata
        else:
            metadata = self.variables[var_index].metadata
        qa_ = self.get_qa_by_number(qa_number, var_index)
        # set:
        if 'qa' not in metadata or metadata.qa is None:
            metadata.qa = []
        if qa_ not in metadata.qa:
            # qa_ is a new element, set values and append
            # add the current qQA nuber and the current key/value pair:
            qa_.qa_number = qa_number
            qa_[key] = value
            # add the new QA element to the list
            metadata.qa.append(qa_)
        else:
            qa_[key] = value

    def unset_qa(self, key, qa_number, var_index):
        """
        Removes a metadata value for a QA metadata element.
        Unset in case of a variable is different from setting to None!
        Parameters:
            key           metadata key
            qa number     number of QA block
            var_index     >0 if variable specific (= from VNAME),
                          None if global (from NCOM block)
        Returns:
            None
        """
        if var_index is None:
            metadata = self.metadata
        else:
            metadata = self.variables[var_index].metadata
        qa_ = self.get_qa_by_number(qa_number, var_index)
        # unset:
        if 'qa' in metadata or metadata.qa is not None and key in metadata.qa:
            del qa_[key]

    def values_without_flags(self, vnum):
        """
        Modifies the values of a timeseries in case it should be used
        without flag information.
        Parameters:
            vnum   variable index
        Modifications:
            values with invalid flag -> MISSING
            values with flag 781 -> value/2.0
        This will maybe hgave to be generalized (moved out of the nasa_ames
        module when other output formats are implemented.
        """
        ret = []
        values = self.variables[vnum].values_
        flags = self.variables[vnum].flags
        for i in range(len(values)):
            (_, invalid, missing, _) = get_flag_summary(flags[i])
            if invalid or missing:
                ret.append(None)
            elif 781 in flags[i] and values[i]:
                ret.append(values[i]/2.0)
            else:
                ret.append(values[i])
        return ret

    def find_variables(self, meta):
        """
        Finds variables in the file which match certain metadata criteria.
        Parameters:
            meta      metadata elements for lookup (dict)
        Returns:
            generator (variable numbers that match)
        """
        for i in range(len(self.variables)):
            if meta:
                for elem in meta.keys():
                    skip = False
                    if elem == 'characteristics':
                        # special case, characteristics is a dict
                        char = self.get_characteristics_for_var(i)
                        for k in meta[elem].keys():
                            # If False in criteria, then the characteristic
                            # may not exist
                            # If True in criteria, then the characteristic must
                            # exist.
                            # Else, the characterisic must be equal the set
                            # value.
                            if ((meta[elem][k] is False and  k in char) or
                                (meta[elem][k] is True and  k not in char) or
                                ((meta[elem][k] is not False and
                                  meta[elem][k] is not True and
                                  (k not in char or meta[elem][k] != char[k])))):
                                skip = True
                                break
                        if skip:
                            break
                    else:
                        if meta[elem] != self.get_meta_for_var(i, elem):
                            skip = True
                            break
                if not skip:
                    yield i

    def find_variable(self, meta):
        """
        Finds exactly one specific variable in the file which matches certain
        metadata criteria.
        Parameters:
            meta      metadata elements for lookup (dict)
        Returns:
            variable number
        Raises:
            ValueError if no variable or more than one found
        """
        itr = self.find_variables(meta)
        try:
            ret = next(itr)
        except StopIteration:
            raise ValueError("no Variable matches criteria")
        try:
            next(itr)
        except StopIteration:
            return ret
        raise ValueError("more than one Variable match criteria")

    def match_variable(self, vnum, meta):
        """
        Check if variable vnum matches criteria meta.
        Parameters:
            vnum     variable number in file
            meta     metadata elements for lookup (dict)
        Returns:
            bool
        """
        if vnum in self.find_variables(meta):
            return True
        return False

    def find_sample_time_index(self, time):
        """
        Finds indexes in sample times that cover a given time or interval.
        Parameters:
            time    datetime.datetime - finding coverage for a point in time
                       or
                    (min, max) - finding coverage for an interval
        Returns:
            (min_index, max_index) or None
        """
        if isinstance(time, datetime.datetime):
            time = DatetimeInterval(time,
                                    time+datetime.timedelta(microseconds=1))
        elif not isinstance(time, DatetimeInterval):
            time = DatetimeInterval(*time)
        time_index = [i for i in range(len(self.sample_times))
                      if time.overlap(self.sample_times[i])]
        if time_index:
            return (time_index[0], time_index[-1])
        return None

    def msgref_vnum_lnum(self, vnum):  # pylint: disable=R0201
        # R0201: Method could be a function
        # --> will be overridden as an instance method.
        """
        Get info on variable number and line number for error messages.
        Default implementation, MUST be overridden by concrete file read
        classes.
        Default returns vnum as variable number an 1 as offset line number.
        Parameters:
            vnum    variable index in the file object
        Returns:
            (file_vnum, file_num) where:
                file_vnum: the real variable number in the physical file
                           (if there was a physical file read)
                           Variable 1 is the first variable after end_time
                file_lnum: the line number offset for the first data line in
                           the physical file (if there was a physical file read)
                file_metalnum: the line number reference that should be used for
                               metadata related messages. Default is None.
        """
        return (vnum+1, 1, None)

    def write_indexdb(self):
        """
        Writes indexdb entries for the io object.
        """
        if self.indexdb:
            self.indexdb.insert_file_obj(self)

    def uniform_time(self):
        """
        Checks if the timeseries is uniform.
        TU (uniform) if the same time difference between the start times.
        (Non necessarily the same as DX. DX is Nasa Ames (i.e. rounded float)
        DX can be != 0 if all differences (rounded to ndigits) are equal DX +-
        1 in the last digit (allow for rounding) 
        Parameters:
            None
        Returns:
            datetime.timedelta object if uniform
            timedelta(0) if not uniform
            None if less than 2 samples.
        """
        totaldiff = None
        for (time1, time2) in zip([time[0] for time in self.sample_times],
                                  [time[0] for time in self.sample_times][1:]):
            diff = time2 - time1
            if totaldiff and diff != totaldiff:
                return datetime.timedelta(0)
            elif not totaldiff:
                totaldiff = diff
        return totaldiff

    def title(self):
        """
        Get a human readable title for the file (depending on parameters and
        station.
        Parameters:
            None
        Returns:
            Title string
        """
        return title(
            [(self.get_meta_for_var(i, 'regime'), self.get_meta_for_var(i, 'matrix'), self.get_meta_for_var(i, 'comp_name'))
             for i in range(len(self.variables))],
            [self.metadata.station_name])

    def set_citation(self, metadata_options):
        """
        Calculate and set the citation metadata from other metadata.
        Parameters:
            metadata_options
                       options for output (bitfield):
        Returns:
            None
        Sets:
            citation
        """
        if not self.from_domain:
            # This file is not generated from the ebas domain.
            # Do not generate a citation string (e.g. submitted files to ebas)
            return
        # make sure citation is not set in any variable:
        for var in self.variables:
            if 'citation' in var.metadata:
                del var.metadata['citation']
        # Authors: join all data originators
        self.metadata.citation = ', '.join(
            ['{}{}'.format(
                ps.PS_LAST_NAME,
                (', ' + ps.PS_FIRST_NAME[0] + '.') if ps.PS_FIRST_NAME else '')
             for ps in self.metadata.originator])
        # Authors: join all project acronyms
        if self.metadata.projects:
            prj = ', '.join(self.metadata.projects)
        else:
            # This means, either different variable have different project lists
            # (can only happen when manually creating the file in ebas-io)
            # or, we are are a joined file, and the project list varies between
            # different metadata_intervals.
            # Both cases are handlesd well by union_projects.
            prj = ', '.join(self.union_projects())
        if self.metadata.citation and prj:
            self.metadata.citation += ', '
        self.metadata.citation += prj
        # Year of publication: from revision date
        if isvalid(self.metadata.revdate):
            self.metadata.citation += ', {}'.format(self.metadata.revdate.year)
        self.metadata.citation += ', ' + self.title()
        self.metadata.citation += ', data hosted by EBAS at NILU'
        if self.metadata.doi:
            self.metadata.citation += ', DOI: {}'.format(
                self.metadata.doi[0])
        elif self.metadata.doi_list:
            self.metadata.citation += ', Part of DOI: {}'.format(
                ', '.join([x for x in self.metadata.doi_list]))

    def union_projects(self):
        """
        Get a union of all projects in the file.
        (Overridden in EbasJoinedFile for adequate handling of joined files)
        Parameters:
            None
        Returns:
            list of project acronyms
        """
        return sorted(
            {prj
             for vnum in range(len(self.variables))
             for prj in self.get_meta_for_var(vnum, 'projects')})

    def set_timespec_metadata(self):
        """
        Calculate and set all time specific metadata from sample times.
        Returns:
            None
        Sets:
            type TU/TI
            timeref, timeoffset
            startdate
        """
        # make sure none of the metadata set here is set in any variable:
        reset_meta = ['period', 'type', 'timeref', 'startdate']
        for var in self.variables:
            for meta in reset_meta:
                if meta in var.metadata:
                    del var.metadata[meta]

        res = estimate_resolution_code(self.sample_times)
        if res or not self.metadata.resolution:
            # set resolution code only if it can be calculated, else keep
            # manual setting if there is one
            self.metadata.resolution = res
        # set period code
        self.metadata.period = estimate_period_code(self.sample_times[0][0],
                                                    self.sample_times[-1][1])
        # set type (TU/TI)
        diff = self.uniform_time()
        if diff is None:
            # Less than 2 samples, type is undefined.
            # If types was set already, leave it (user might want this)
            # If nothing is set, set TI
            if self.metadata.type not in ("TU", "TI"):
                self.metadata.type = "TI"
        elif diff == datetime.timedelta(0):
            self.metadata.type = "TI"
        else:
            self.metadata.type = "TU"

        # do not use timeref on export anymore!
        self.metadata.timeref = None
        self.internal.timeoffset = datetime.timedelta(seconds=0)

        self.metadata.startdate = self.sample_times[0][0]

    def join(self, other):  # @UnusedVariable  pylint: disable=W0613, R0201
        # UnusedVariable: only in default implementation
        # W0613: Unused argument 'other': only in default implementation
        # R0201: Method could be a  function: only in default implementation
        """
        Join self with other if possible.
        Default implementation: do not join
        Parameters:
            other    other EbasIO object
        Returns
            New, joined EbasIO object
            None (no join possible)
        """
        return None

    def check(  # pylint: disable=R0913
            # R0913  Too many arguments
            self, ignore_rescode=False, ignore_revdate=False,
            ignore_parameter=False, ignore_sampleduration=False,
            ignore_identicalvalues=False,
            ignore_flagconsistency=False, fix_flagconsistency=False,
            ignore_nodata=False,
            ignore_valuecheck=False, ignore_templatecheck=False,
            fix_overlap_set_start=False,
            fix_bc211=False):
        """
        Performs a full check of the IO object.
        Parameter:
            ignore_rescode    ignore rescde errors (downgraded to warning)
            ignore revdate    ignore revdate errors (downgrade to warning)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
            ignore_identicalvalues
                              ignore errors related to all values of one
                              variable being equal (e.g. all data missing for
                              one variable, all values 0.0 or all values any
                              other constant value).
            ignore_flagconsistency
                              ignore errors regarding flag consistency and
                              flag/value consitency
            fix_flagconsistency
                              ignore errors regarding flag consistency and
                              flag/value consitency; additionally fix problems
                              if possible
            ignore_nodata     the object is expected to not contain any data
                              (ie.g. read with skip_data)
            ignore_valuecheck ignore value checks (boundaries and spikes)
            ignore_templatecheck ignore template checks
            fix_overlap_set_start
                              instead of raising an error in case of overlaps,
                              set the previous endtime to start time
            fix_bc211         flag boundary check violations with flag 211
        """
        self.check_metadata(
            ignore_rescode=ignore_rescode, ignore_revdate=ignore_revdate,
            ignore_parameter=ignore_parameter,
            ignore_sampleduration=ignore_sampleduration)
        self.check_data(ignore_identicalvalues=ignore_identicalvalues,
                        ignore_nodata=ignore_nodata,
                        ignore_flagconsistency=ignore_flagconsistency,
                        fix_flagconsistency=fix_flagconsistency,
                        ignore_valuecheck=ignore_valuecheck,
                        fix_overlap_set_start=fix_overlap_set_start,
                        fix_bc211=fix_bc211)
        try:
            template = get_template(
                self, ignore_templatecheck=ignore_templatecheck)
        except NoTemplate:
            self.warning('No applicable template found')
        else:
            self.logger.info("found template {}".format(template.TEMPLATE_NAME))
            template.checkfile()

    def check_metadata(
            self, ignore_rescode=False, ignore_revdate=False,
            ignore_parameter=False):
        """
        Abstract in the base class, implemented in EbasFileCheckMetadata.
        Performs a check of the metadata of the object.
        Parameter:
            ignore_rescode    ignore rescde errors (downgraded to warning)
            ignore revdate    ignore revdate errors (downgrade to warning)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
        """
        raise NotImplementedError

    def check_data(self, ignore_identicalvalues=False,
                   ignore_flagconsistency=False, fix_flagconsistency=False,
                   ignore_nodata=False,
                   ignore_valuecheck=False, fix_overlap_set_start=False,
                   fix_bc211=False):
        """
        Abstract in the base class, implemented in EbasFileCheckData.
        Check data consistency and any cause for suspicion in the data.
        Parameter:
            ignore_identicalvalues
                    ignore errors related to all values of one
                    variable being equal (e.g. all data missing for
                    one variable, all values 0.0 or all values any
                    other constant value).
            ignore_flagconsistency
                    ignore errors regarding flag consistency and flag/value
                    consitency
            ignore_nodata
                    the object is expected to not contain any data (ie.g. read
                    with skip_data)
            ignore_valuecheck ignore value checks (boundaries and spikes)
            fix_overlap_set_start
                    instead of raising an error in case of overlaps,
            fix_bc211         flag boundary check violations with flag 211
        """
        raise NotImplementedError

    def update_metadata(self, meta, vnum=None):
        """
        Merges the fileglobal or variable specific metadata with a dictionary.
        !!! This merges metadata key wise, e.g. no special trearment for "qa".
        !!! --> "qa" is taken as a whole.
        Parameters:
            vnum    variable number (if variable specific metadata)
            meta    dictionary with metadata to be updated
        Returns:
            None
        """
        if vnum is None:
            mymeta = self.metadata
        else:
            mymeta = self.variables[vnum].metadata
        for key in meta.keys():
            if key not in mymeta or mymeta[key] != meta[key]:
                # updat needed
                if isinstance(meta[key], dict):
                    # Make DataDict from all dict like objects
                    mymeta[key] = DataObject(meta[key])
                elif isinstance(meta[key], list):
                    # Make list of Datadict from all list of dict like objects
                    mymeta[key] = list()
                    for elem in meta[key]:
                        if isinstance(elem, dict):
                            mymeta[key].append(DataObject(elem))
                        else:
                            mymeta[key].append(elem)
                else:
                    mymeta[key] = meta[key]

    def get_all_meta_for_var(self, vnum):
        """
        Get all metadtaa for one variable (merge global and variable specific).
        """
        res = DataObject()
        for key in set(self.metadata.keys() + \
                       self.variables[vnum].metadata.keys()):
            if key == 'qa':
                res[key] = [x for x in self.get_qa_for_var(vnum)]
            else:
                res[key] = self.get_meta_for_var(vnum, key)
        return res
        
        



"""
notification.py
$Id: notification.py 2452 2020-04-23 19:20:47Z pe $

General notification / alert handling.
Collects messages and sends emails with collected messages at demand.
As a fallback (mail sending problems) notifications will be written to logfile.

History:
V.1.0.0  2013-06-26  pe  initial

"""

import smtplib
import socket
import logging

class NotificationDeliveryError(Exception):
    """
    Exception raised on email delivery errors.
    """
    pass

class Notification(object): # pylint: disable=R0902
    # R0902: Too many instance attributes
    """
    Class for keeping info on all notification during the processing of one
    file.
    """
    def __init__(self, smtp, sender="", fromname="", recipients=None,
                 bcc=None, subject=None, headertext='', trailertext='',
                 override_recipients=None, parentlogger=logging.getLogger()):
        # pylint: disable=R0913
        # R0913: Too many arguments
        """
        Set up notification object.
        """
        self.messages = []
        self.smtp = smtp
        self.sender = sender
        self.fromname = fromname
        if isinstance(recipients, (list, tuple)):
            self.recipients = list(recipients)
        elif  recipients:
            self.recipients = [recipients]
        else:
            self.recipients = []
        if isinstance(bcc, (list, tuple)):
            self.bcc = list(bcc)
        elif  bcc:
            self.bcc = [bcc]
        else:
            self.bcc = []
        if isinstance(override_recipients, (list, tuple)):
            self.override_recipients = list(override_recipients)
        elif  override_recipients:
            self.override_recipients = [override_recipients]
        else:
            self.override_recipients = []
        self.subject = subject
        self.headertext = headertext
        self.trailertext = trailertext
        self.logger = parentlogger.getChild('notifications')

    def add(self, message):
        """
        Add a message to the data submitter(s).
        """
        self.messages.append(message)

    def send(self):
        """
        Send all pending notifications.
        """
        if not self.messages:
            return
        self.logger.info("sending pending notification email")

        try:
            smtp = smtplib.SMTP(self.smtp, timeout=5)
        except (socket.error, socket.timeout, socket.gaierror,
                smtplib.SMTPConnectError) as expt:
            errtxt = "error connecting to mailserver " + self.smtp
            errtxt += "; " + str(expt)
            raise NotificationDeliveryError(errtxt)

        msg = self.generate_body()
        if self.override_recipients:
            rec = self.override_recipients
        else:
            rec = self.recipients + self.bcc

        try:
            smtp.sendmail(self.sender, rec, msg)
        except (smtplib.SMTPHeloError, smtplib.SMTPRecipientsRefused,
                smtplib.SMTPSenderRefused, smtplib.SMTPDataError) as expt:
            errtxt = "send email failed; {}: {}".format(expt.__class__.__name__,
                                                        str(expt))
            self.logger.error(errtxt)
            raise NotificationDeliveryError(errtxt)
        smtp.quit()

    def generate_body(self):
        """
        Generate email body.
        Parameters:
            None
        Returns:
            None
        """
        msg = "From: {} <{}>\r\n".format(self.fromname, self.sender) +\
              "To: {0}\r\n".format(", ".join(self.recipients)) +\
              "Subject: {}\r\n\r\n".format(self.subject) +\
              self.headertext +\
              '\r\n'.join(self.messages) +\
              self.trailertext
        if self.override_recipients:
            msg = "From: {} <{}>\r\n".format(self.fromname, self.sender) +\
                  "To: {0}\r\n".format(", ".join(self.override_recipients)) +\
                  "Subject: {}\r\n\r\n".format(self.subject) +\
                  "Email Notification - " +\
                  "Addressee overwritten\r\n\r\n" +\
                  "=== original email follows ===\r\n\r\n" +\
                  msg
        return msg



"""
$Id: read_precip.py 2770 2021-12-08 22:36:25Z pe $

File input functionality for EbasEanetPrecip
"""

import datetime
from nilutility.datatypes import DataObject, recursive_data_object
from ebas.domain.basic_domain_logic.flags import get_flag_summary
from fileformats.EANET import EanetPrecip, EanetError
from .eanet_cfg import EANET_SITES, EANET_PRECIP_ANALYTICAL_LABS, \
    EANET_PRECIP_REPORTERS, EANET_PRECIP_GLOBAL, EANET_PRECIP_PARAMS, \
    EANET_PRECIP_DETAIL
from .base import EbasEanetPrecipBase, EbasEanetReadError

EANET_SITES = recursive_data_object(EANET_SITES)
EANET_PRECIP_ANALYTICAL_LABS = recursive_data_object(
    EANET_PRECIP_ANALYTICAL_LABS)
EANET_PRECIP_REPORTERS = recursive_data_object(EANET_PRECIP_REPORTERS)
EANET_PRECIP_GLOBAL = recursive_data_object(EANET_PRECIP_GLOBAL)
EANET_PRECIP_PARAMS = recursive_data_object(EANET_PRECIP_PARAMS)
EANET_PRECIP_DETAIL = recursive_data_object(EANET_PRECIP_DETAIL)


class EanetPrecipPartialRead(EbasEanetPrecipBase):
    """
    Read fuctionality for EbasEanetPrecip I/O object.
    """

    UNIQUE_ID = 0

    def get_unique_id(self):
        """
        Get a unique int id
        """
        self.__class__.UNIQUE_ID += 1
        return self.__class__.UNIQUE_ID

    def _read(self, filename, skip_unitconvert=False, **args):
        """
        Reads a EANET precip file into an EBAS I/O object.
        Parameters:
            filename          path and filename
            other paramters are ignored: ignore_rescode, ignore revdate,
                ignore_parameters, skip_data, skip_unitconvert, skip_variables,
                encoding, ignore_numformat

        Returns:
            None
        Raises:
            IOError (from builtin, file open)
            EbasEanetReadError
        """
        # EANET precip file has no revdata metadata, use now:
        self.metadata.revdate = datetime.datetime.utcnow()
        self.sourcefile = EanetPrecip()
        try:
            self.sourcefile.read(filename)
        except EanetError:
            pass
        self.errors += self.sourcefile.errors
        self.warnings += self.sourcefile.warnings
        if self.errors:
            self.logger.info("Exiting because of previous errors")
            raise EbasEanetReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))
        self.parse()
        if self.errors:
            self.logger.info("Exiting because of previous errors")
            raise EbasEanetReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))
        self.prepare_write(False)
        if self.errors:
            self.logger.info("Exiting because of previous errors")
            raise EbasEanetReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))

    def parse(self):
        """
        Parse all metadata and data from the source file object to the ebas-io
        object.
        """
        self._prec_values = None

        global_meta = EANET_PRECIP_GLOBAL.EBAS.copy()
        global_eanet_meta = DataObject()
        global_eanet_meta.update(EANET_PRECIP_GLOBAL.EANET)
        try:
            global_eanet_meta.update(
                EANET_SITES[self.sourcefile.site_code]['EANET'])
            global_meta.update(EANET_SITES[self.sourcefile.site_code]['EBAS'])
        except KeyError:
            self.error("Site '{}' not found in configuration".format(
                self.sourcefile.site_code))
            raise EbasEanetReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))
        if self.sourcefile.site_name != global_eanet_meta.site_name:
            self.error(
                "Site name mismatch: file '{}' inconsistent with "
                "configuration".format(self.sourcefile.file_name))

        self._map_analytical_lab(global_meta)
        self._map_person(global_meta)

        self.sample_times = self.sourcefile.sample_times
        for var_key, conf in EANET_PRECIP_PARAMS.items():
            # first round, only amount
            if conf.EBAS.comp_name.startswith('precipitation_amount'):
                self._add_var(var_key, global_meta, global_eanet_meta)
        for var_key, conf in EANET_PRECIP_PARAMS.items():
            # second round, all other variables
            if not conf.EBAS.comp_name.startswith('precipitation_amount'):
                self._add_var(var_key, global_meta, global_eanet_meta)

    def _add_var(self, var_key, global_meta, global_eanet_meta):
        """
        Convert a single variable.
        Parameters:
            var_key            the variable key in the eanet configuration
            global_meta        global metadata
            global_eanet_meta  global eanet metadata
        Returns:
            tuple (metadata, values, flags)
        """
        metadata = global_meta.copy()
        eanet_meta = global_eanet_meta.copy()
        metadata.update(EANET_PRECIP_PARAMS[var_key].EBAS)
        eanet_meta.update(EANET_PRECIP_PARAMS[var_key].EANET)

        # Update/override metadata from EANET_PRECIP_DETAIL configuration
        try:
            EANET_PRECIP_DETAIL[self.sourcefile.site_code]
        except KeyError:
            self.error(
                "Site '{}' not found in EANET_PRECIP_DETAIL "
                "configuration".format(self.sourcefile.site_code))
            return
        else:
            try:
                conf = EANET_PRECIP_DETAIL[self.sourcefile.site_code][var_key]
            except KeyError:
                conf = EANET_PRECIP_DETAIL[self.sourcefile.site_code][None]
            metadata.update(conf.EBAS)
            eanet_meta.update(conf.EANET)

        self._set_station(metadata)
        self._set_org(metadata)
        self._set_method(metadata)

        form_number = eanet_meta.form_number
        var_name = eanet_meta.variable_name

        var = self.sourcefile.forms[form_number].var_by_name(var_name)

        if var.unit != eanet_meta.unit:
            self.error(
                "{} ({}): EANET unit mismatch between file and "
                "configuration".format(var_name, var_key))
            return

        if not [x for x in var.samples if x is not None]:
            self.warning("{} ({}): no valid values; skipping".format(
                var_name, var_key))
            return

        flags = var.flags
        if flags is None:
            # This variable has no flags in the EANET file. Set empty flags.
            flags = [[] for _ in range(len(var.samples))]
        # add some missing flags:
        for i, val in enumerate(var.samples):
            if val is None:
                _, _, missing = get_flag_summary(flags[i], want_issues=False)
                if self._prec_values and self._prec_values[i] == 0.0 and \
                        890 not in flags[i]:
                    # no precipitation: add flag 890
                    flags[i].append(890)
                elif not missing:
                    # value is missing, but no missing flag set:
                    flags[i].append(999)
        if not self._prec_values and \
                metadata.comp_name.startswith('precipitation_amount'):
            self._prec_values = var.samples
        self._add_variable_nondomain(
            self.get_unique_id(), metadata, var.samples, flags, 0)

    def _map_analytical_lab(self, metadata):
        """
        Map the analytical lab code.
        """
        # override #ana_lab_code even when it is set in station config
        # needed for precip mapping (filter needes #ana_lab_code from station
        # config, but precip uses this mapping.
        # If setting in config is needed, set in EANET_PRECIP_DETAIL
        try:
            metadata['#ana_lab_code'] = EANET_PRECIP_ANALYTICAL_LABS[
                self.sourcefile.laboratory]
        except KeyError:
            self.error("Organisation '{}' not found in configuration".format(
                       self.sourcefile.laboratory))
            metadata['#ana_lab_code'] = ''

    def _map_person(self, metadata):
        """
        Map originater and submitter from reporter
        """
        nobody = DataObject({
            'PS_LAST_NAME': None,
            'PS_FIRST_NAME': None,
            'PS_EMAIL': None,
            'PS_ORG_NAME': None,
            'PS_ORG_ACR': None,
            'PS_ORG_UNIT': None,
            'PS_ADDR_LINE1': None,
            'PS_ADDR_LINE2': None,
            'PS_ADDR_ZIP': None,
            'PS_ADDR_CITY': None,
            'PS_ADDR_COUNTRY': None,
            'PS_ORCID': None,
            })
        if self.sourcefile.reporter not in EANET_PRECIP_REPORTERS:
            self.warning(
                "Repoorter '{}' not found in configuration, "
                "using as is.".format(self.sourcefile.reporter))
            person = nobody.copy()
            person.update({'PS_LAST_NAME': self.sourcefile.reporter})
            pers_list = [person]
        else:
            pers_list = []
            for pers in EANET_PRECIP_REPORTERS[self.sourcefile.reporter]:
                person = nobody.copy()
                person.update(pers)
                pers_list.append(person)
        metadata['originator'] = pers_list
        metadata['submitter'] = pers_list



"""
$Id: msg_condenser.py 2427 2020-03-19 23:09:09Z pe $

Message condenser, a way to deal with many repeating messages (e.g. logging).

= Synopsis:

from nilutility.msg_condenser import MessageCondenser

msgc = MessageCondenser(10)

def printit(*args):
    print args[0]

for i in range(1000000):
    msgc.add(printit, 1, 'message')
msgc.deliver()

==>

message
The previous message repeats 999999 times


Both classes will for real world scenarios be base classes for more specialized
derived classes. See fileformats/NasaAmes and ebas.io.file/basefile for
examples.

"""

from collections import defaultdict

class MessageRecord(tuple):
    """
    Message record class for MessageCondenser
    """

    KEYS = ['msgfunc', 'msg_id']
    DATA = ['message']

    def __getattr__(self, name):
        """
        Get key and data elements as attributes.
        Parameters:
            name    attribute name
        Returns:
            tuple value which corresponds to the respective key or data element.
        Raises:
            Attribute error if attribute name does not exists in key or data.
        """
        if name in self.__class__.KEYS:
            return self[self.__class__.KEYS.index(name)]
        if name in self.__class__.DATA:
            return self[len(self.__class__.KEYS) +
                        self.__class__.DATA.index(name)]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            self.__class__.__name__, name))

    @property
    def key(self):
        """
        Get the key values of the record as tuple.
        Parameters:
            None
        Returns:
            Key values
        """
        return self[0:len(self.__class__.KEYS)]

class MessageCondenser(object):
    """
    Condense recurring messages.
    """

    def __init__(self, threshold=10, record_cls=MessageRecord):
        """
        Set up condenser object.
        Parameters:
            threshold   message threshold (from this number of occurrences the
                        messages will be condensed.
                        0 -> never condense
        Returns:
            None
        """
        self.threshold = threshold
        self.record_cls = record_cls
        self.list = []
        self.dict = defaultdict(list)

    def add(self, *args):
        """
        Add a message to the object.
        Parameter:
            *args: all keys and all data elements (additional elements included)
        Returns:
            None
        """
        record = self.record_cls(args)
        self.list.append(record)
        self.dict[record.key].append(record)

    def deliver(self):
        """
        Deliver the messages.
        """
        processed = []
        for msg in self.list:
            key = msg.key
            if key in processed:
                continue
            if self.threshold != 0 and len(self.dict[key]) >= self.threshold:
                # condense messages
                self.deliver_msg(msg, True)
                processed.append(key)
            else:
                # deliver a single message
                self.deliver_msg(msg, False)
        self.reset()

    def deliver_msg(self, record, condensed):  # pylint: disable=R0201
        # R0201: Method could be a  function
        # --> this is usually overridden with a real method...
        """
        Deliver a single message.
        Parameters:
            record
        Returns:
            None
        """
        record.msgfunc(record.message)
        if condensed:
            record.msgfunc(
                "The previous message repeats {} times".format(
                    len(self.dict[record.key])-1))

    def reset(self):
        """
        Reset the object to a freshly initialized state.
        """
        self.list = []
        self.dict = defaultdict(list)

    def is_reset(self):
        """
        Check status if object is reset.
        Parameters:
            None
        Returns:
            True if reset, else False
        """
        if self.list or self.dict:
            return False
        return True



"""
ebas/time_period.py
$Id: time_period.py 2794 2022-04-06 14:41:36Z pe $

functions for time intervals period/resolution code

History:
V.1.0.0  2013-05-30  pe  initial version

"""

import re
import dateutil.relativedelta
import datetime
from nilutility.statistics import modified_median
from nilutility.datetime_helper import DatetimeISO8601Duration

GREGORIAN_DAYS = 365.+(1./4)-(1./100)+(1./400)

PERIOD_CODE_FACTORS = {
                     'ms':  0.001,
                     's':  1,
                     'mn': 60,
                     'h':  60*60,
                     'd':  60*60*24,
                     'w':  60*60*24*7,
                     'mo': 60*60*24*(GREGORIAN_DAYS)/12.,
                     'y': 60*60*24*(GREGORIAN_DAYS)
    # month:    ...*(days_per_year)/12
    #  average days per year according to gregorian calendar,
    #  ignoring leap seconds
                     }

def parse_period_code(rescode):
    """
    Parse the period code string.
    Parameters:
        rescode    period code string
    Returns:
        tupel (value, unit) where value is the integer number and unit is the
        time unit
    Raises:
        ValueError where the string can not be parsed
    """
    match = re.match(r'^(\d+)(ms|s|mn|h|d|w|mo|y)$', rescode)
    if not match:
        raise ValueError("illegal period code {0}".format(rescode))
    return (int(match.group(1)), match.group(2))

def period_code_seconds(rescode):
    """
    Returns the given periode in converted to seconds.
    This is an estimate for month ('mo' periode codes), else it should be
    exact.
    e.g. '2w'  --> 2*7*24*60*60 = 1209600.0
         '6mo' --> 6*30.4*24*60*60 = 15759360.0
         '5mn' --> 5*60 = 300.0
         '400ms' --> 0.4
    Parameters:
        rescode    resolution code string, number+time unit
                   (time units: s, mn, h, d, w, mo)
    Returns:
        rescode converted to seconds
    Raises:
        ValueError
            when period code not known
    """
    (value, unit) = parse_period_code(rescode)
    return value * PERIOD_CODE_FACTORS[unit]

def period_code_iso8601_duration(rescode):
    """
    Returns the given periode as a DatetimeISO8601Duration object.
    
    e.g. '2w'  --> P14d or P0000-0014T00:00:00
         '6mo' --> P6M or P0000-06-00T00:00:00
         '5mn' --> PT5M or P0000-00-00T00:05:00
         '400ms' --> PT0.4S or P0000-00-00T00:00:00.400000
    Parameters:
        rescode    resolution code string, number+time unit
                   (time units: s, mn, h, d, w, mo)
    Returns:
        DatetimeISO8601Duration object
    """
    dur = DatetimeISO8601Duration()
    (value, unit) = parse_period_code(rescode)
    if value == 0:
        # 0 whatever unit is quite undefined... return empty duration
        return dur
    if unit == 'y':
        dur.year = value
    if unit == 'mo':
        dur.month = value
    if unit == 'w':
        dur.day = value * 7
    if unit == 'd':
        dur.day = value
    if unit == 'h':
        dur.hour = value
    if unit == 'mi':
        dur.minute = value
    if unit == 's':
        dur.second = value
    if unit == 'ms':
        dur.microsecond = value * 1000
    return dur

def normalize_period_code(rescode):
    """
    Returns the given periode in ebas standard period code.
    (which is the shortest possible integer number)
    e.g.:
         '300s'    --> '5mn'
         '60mn'    --> '1h'
         '3600s'   --> '1h'
         '48h'     --> '2d'
         '7d'      --> '1w'
         '604800s' --> '1w'
    Month and year codes will not be generated (both are variable length).
    
    Parameters:
        rescode    resolution code string, number+time unit
                   (time units: s, mn, h, d, w, mo)
    Returns:
        rescode converted to standard
    Raises:
        ValueError
            when period code not known
    """
    (value, unit) = parse_period_code(rescode)
    if value == 0:
        # 0 whatever unit is quite undefined... return original unit
        return rescode
    if unit == 'ms' and value >= 1000:
        value /= 1000
        unit = 's'
    if unit == 's' and value % 60 == 0:
        value /= 60
        unit = 'mn'
    if unit == 'mn' and value % 60 == 0:
        value /= 60
        unit = 'h'
    if unit == 'h' and value % 24 == 0:
        value /= 24
        unit = 'd'
    if unit == 'd' and value % 7 == 0:
        value /= 7
        unit = 'w'
    return str(int(value))+unit

def estimate_period_code(start, end):
    # pylint: disable-msg=R0911, R0912
    # R0912: Too many branches
    # R0911: Too many return statements 
    """
    Returns the closest approximation of a periode code.
    e.g.:
    
    Parameters:
        start    datetime start of interval
        end      datetime end of interval
    Returns:
        rescode
        None if difference is < 0.0005 sec
    """
    tot_sec = (end - start).total_seconds()
    
    y_real = tot_sec / (GREGORIAN_DAYS*24*60*60.0)
    y_est = int(round(y_real))
    # if exact y --> return y
    if y_est and \
       abs((end-(start+dateutil.relativedelta.relativedelta(years=y_est)))\
       .total_seconds()) < 1.0:
        return '{}y'.format(y_est)
    # if > 3y --> return rounded y
    if y_est >= 3:
        return '{}y'.format(y_est)
    
    # if exact month --> return mo
    mo_est = int(round(tot_sec / (GREGORIAN_DAYS*24*60*60.0/12.0)))
    if mo_est and \
       abs((end-(start+dateutil.relativedelta.relativedelta(months=mo_est)))\
           .total_seconds()) < 1.0:
        return '{}mo'.format(mo_est)
    
    # if exact week --> return w
    w_real = tot_sec / (7*24*60*60.0)
    w_est = int(round(w_real)) 
    if w_est and abs(tot_sec - w_est * 7*24*60*60.0) < 1.0:
        return '{}w'.format(w_est)
    
    # 2 y +- 3w -> rounded y
    if y_est == 2 and \
       abs((end-(start+dateutil.relativedelta.relativedelta(years=y_est)))\
           .total_seconds()) < (3*7*24*60*60.0):
        return '{}y'.format(y_est)
    # 1 y +- 1w -> rounded y
    if y_est == 1 and \
       abs((end-(start+dateutil.relativedelta.relativedelta(years=y_est)))\
           .total_seconds()) < (7*24*60*60.0):
        return '{}y'.format(y_est)
    
    # if >= 2y --> round mo
    if y_est >= 2:
        return '{}mo'.format(mo_est)

    # mo +-0.5d --> round mo
    if mo_est and \
       abs((end-(start+dateutil.relativedelta.relativedelta(months=mo_est)))\
           .total_seconds()) <= (12*60*60.0):
        return '{}mo'.format(mo_est)

    # w +-0.5d --> round w
    if w_est and abs(tot_sec - w_est * 7*24*60*60) <= (12*60*60.0):
        return '{}w'.format(w_est)

    # mo >= 3 and +-1d --> round mo
    if mo_est >= 3 and \
       abs((end-(start+dateutil.relativedelta.relativedelta(months=mo_est)))\
           .total_seconds()) <= (24*60*60.0):
        return '{}mo'.format(mo_est)

    # w >= 12 and +-1d --> round w
    if w_est >= 12 and abs(tot_sec - w_est * 7*24*60*60) < (24*60*60.0):
        return '{}w'.format(w_est)
    
    # mo >= 3 --> round mo
    if mo_est >= 3:
        return '{}mo'.format(mo_est)
        
    d_real = tot_sec / (24*60*60.0)
    d_est = int(round(d_real))
    # exact d --> return d
    if d_est and abs(tot_sec - d_est * 24*60*60.0) < 1.0:
        return '{}d'.format(d_est)
    
    # w >= 3 and +-2d --> round w
    if w_est >= 3 and abs(tot_sec - w_est * (7*24*60*60.0)) < (2*24*60*60.0):
        return '{}w'.format(w_est)

    # d >= 4 --> return rounded d
    if d_real > 4.0:
        return '{}d'.format(d_est)

    h_real =  tot_sec / (60*60.0)
    h_est = int(round(h_real))
    # h >= 48 --> return rounded h or    
    # h >= 5 +-5mn --> return rounded h
    # h >= 1 +-30sec --> return rounded h
    if h_est >= 24 or \
       (h_est >= 5 and abs(tot_sec - (h_est * (60*60.0))) < 300.0) or \
       (h_est >= 1 and abs(tot_sec - (h_est * (60*60.0))) < 30):
        return '{}h'.format(h_est)

    mn_real =  tot_sec / (60.0)
    mn_est = int(round(mn_real))
    # exact mn --> return mn    
    if mn_est and abs(tot_sec - mn_est * 60.0) < 1.0:
        return '{}mn'.format(mn_est)
    
    # h >= 5 --> return rounded h    
    if h_est >= 5:
        return '{}h'.format(h_est)
    
    # mn +-5s and minute is n/4h --> return mn    
    if mn_est and abs(tot_sec - mn_est * 60) < 5.0 and mn_est % 15 == 0:
        return '{}mn'.format(mn_est)
    
    # h >= 1 --> return rounded h    
    if h_real >= 1.0:
        return '{}h'.format(h_est)
    
    # mn +-5s or minute > 15 -> return mn
    if mn_est and (abs(tot_sec - mn_est * 60) < 5.0 or mn_est > 15):
        return '{}mn'.format(mn_est)
    
    # if >= 0.5, use s
    if tot_sec >= 0.5:
        return "{}s".format(int(round(tot_sec)))

    # if between 0.5 ms and 500 ms: use ms
    if tot_sec >= 0.0005:
        return "{}ms".format(int(round(tot_sec*1000)))

    return None

def estimate_resolution_code(sample_times):
    """
    Calculates a representative resolution code based on sample start/end times.
    The estimate is in most cases the median of all start time differences in
    data. See implementation of modified_median for details.
    Parameters:
        sample_times   list of tuples [(start, end),...]
    Returns:
        representative resolution code (e.g. 2h)
        None if undefined (e.g. only one sample or any sample time is None)
    """
    if len(sample_times) > 1:
        try:
            med_res = modified_median([diff[1][0] - diff[0][0] for diff in \
                        zip(sample_times[:-1], sample_times[1:])])
        except TypeError:
            # most likely one time is None...
            return None
        res_sec = med_res.total_seconds()
        res_cde = estimate_period_code(datetime.datetime(1900, 1, 1), \
            datetime.datetime(1900, 1, 1) + med_res)
        return res_cde
    return None

def estimate_sample_duration_code(sample_times):
    """
    Calculates a representative sample duration code based on sample start/end
    times.
    The estimate is in most cases the median duration in data. See
    implementation of modified_median for details.
    Parameters:
        sample_times   list of tuples [(start, end),...]
    Returns:
        representative sample duration code (e.g. 2h)
        None if undefined (e.g. only one sample or any sample time is None)
    """
    if len(sample_times) > 0:
        try:
            med_dur = modified_median([samp[1] - samp[0]
                              for samp in sample_times])
        except TypeError:
            # most likely one time is None...
            return None
        dur_sec = med_dur.total_seconds()
        return estimate_period_code(datetime.datetime(1900, 1, 1), \
            datetime.datetime(1900, 1, 1) + med_dur)
    return None



"""
$Id: __init__.py 2749 2021-11-25 00:34:17Z pe $

EBAS I/O EAET Filter module
"""

from .base import EbasEanetReadError
from .eanet_filter import EbasEanetFilter
from .eanet_precip import EbasEanetPrecip
